'''Graph neural network implementation.'''

from typing import Any, Callable

import numpy as np
import torch
from torch import nn, Tensor
from torch.nn import functional as fn
from torch.optim import Optimizer
from torch_geometric.data import Data
from torch_geometric.loader import DataLoader
import torch_geometric.nn as pyg_nn
from tqdm import tqdm

from ..exception import ArgumentError
from .tag import DEPREL_TO_IDX
from .tree import Node, Tree


__all__ = (
    'GAT',
    'GCN',
    'build_pyg_dataset',
    'get_edge_ohe',
    'train_gnn',
)


DataBatch = Any
LRScheduler = Any


def get_edge_ohe(child: Node) -> Tensor:
    '''
    Get one-hot encoded vector for an edge from the current node to its parent.

    :param child: dependant node of a relation
    :return: one-hot encoded torch.Tensor for a type of dependency relation
    '''

    n_deps = len(DEPREL_TO_IDX)
    attr = torch.zeros(n_deps - 1)
    idx = DEPREL_TO_IDX[child.deprel]
    if idx < n_deps - 1:
        attr[idx] = 1
    return attr


def build_pyg_dataset(
    tree_pairs: list[tuple[Tree, Tree]],
    labels: list[int] | None = None,
    get_edge_attr: Callable[[Node], Tensor] = get_edge_ohe,
) -> list[Data]:
    '''
    Get list of torch_geometric.data.Data instances from a list of tree pairs.

    :param tree_pairs: list of tree pairs composing a dataset
    :param labels: list of labels for tree pairs
    :param get_edge_attr: callback for obtaining edge attribute vectors
    :return: list of torch_geometric.data.Data instances
    '''

    dataset = []

    for i, (lhs, rhs) in enumerate(tree_pairs):
        if any(node.embedding is None for tree in (lhs, rhs) for node in tree):
            raise ArgumentError(
                'All tree nodes should have an embedding '
                'to build a dataset for graph NN model.'
            )

        x = torch.cat(
            [node.embedding.reshape(1, -1) for node in lhs] +  # type: ignore
            [node.embedding.reshape(1, -1) for node in rhs],   # type: ignore
            dim=0,
        )

        shift = lhs.size
        edge_index = []
        edge_attr = []
        for node in lhs:
            for child in node.children:
                edge_index.append((node.idx - 1, child.idx - 1))
                attr = get_edge_attr(child)
                edge_attr.append(attr.reshape(1, -1))
        for node in rhs:
            for child in node.children:
                edge_index.append((shift + node.idx - 1, shift + child.idx - 1))
                attr = get_edge_attr(child)
                edge_attr.append(attr.reshape(1, -1))

        edge_attr_tensor = (
            torch.cat(edge_attr, dim=0)
            if edge_attr
            else torch.tensor([], dtype=float)  # type: ignore
        )

        data = Data(
            x=x,
            edge_index=torch.tensor(edge_index, dtype=int).T.contiguous(),  # type: ignore
            edge_attr=edge_attr_tensor,
            y=labels[i] if labels is not None else None,
        )
        dataset.append(data)

    return dataset


class GCN(nn.Module):
    '''Graph convolutional binary classification network implementation.'''

    def __init__(
        self,
        *,
        input_dim: int = 300,
        hidden_dim: int = 64,
        num_layers: int = 1,
        dropout: float = 0,
        edge_dim: int | None = None,
    ):
        '''
        GCN initialization.

        :param input_dim: dimensionality of an input vector
        :param hidden_dim: dimensionality of vectors in hidden layers
        :param num_layers: number of hidden convolutional layers
        :param dropout: dropout rate
        :param edge_dim: dimensionality of edge attribute vectors
        '''

        super().__init__()

        self.dropout = dropout

        self.convs = nn.ModuleList()
        for i in range(num_layers):
            in_dim = hidden_dim if i else input_dim
            self.convs.append(pyg_nn.Sequential('x, edge_index', [
                (pyg_nn.convs.GCNConv(
                    in_dim,
                    hidden_dim,
                    concat=False,
                    edge_dim=edge_dim,
                ), 'x, edge_index -> x'),
                pyg_nn.norm.GraphNorm(in_channels=hidden_dim),
                nn.PReLU(num_parameters=hidden_dim),
            ]))

        self.post_mp = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.Dropout(dropout),
            nn.PReLU(num_parameters=hidden_dim),
            nn.Linear(hidden_dim, 1),
            nn.Sigmoid(),
        )

    def forward(self, data: DataBatch) -> Tensor:
        '''
        Apply a forward pass for a data batch.

        :param data: torch_geometric.data.batch.DataBatch instance to process
        :return: torch.Tensor with class probabilities for each data instance in a batch
        '''

        x, edge_index, batch = data.x, data.edge_index, data.batch

        for conv in self.convs:
            x = conv(x, edge_index)
        x = pyg_nn.global_max_pool(x, batch)
        x = self.post_mp(x)

        return x


class GAT(nn.Module):
    '''Graph attention binary classification network implementation.'''

    def __init__(
        self,
        *,
        input_dim: int = 300,
        hidden_dim: int = 64,
        num_heads: int = 1,
        num_layers: int = 1,
        dropout: float = 0,
        negative_slope: float = 0.2,
        edge_dim: int | None = None,
    ):
        '''
        GAT initialization.

        :param input_dim: dimensionality of an input vector
        :param hidden_dim: dimensionality of vectors in hidden layers
        :param num_heads: number of attention heads
        :param num_layers: number of hidden convolutional layers (per each attention head)
        :param dropout: dropout rate
        :param negative_slope: slope coefficient for LeakyReLU layers of attention heads
        :param edge_dim: dimensionality of edge attribute vectors
        '''

        super().__init__()

        self.dropout = dropout

        self.convs = nn.ModuleList()
        for i in range(num_layers):
            in_dim = hidden_dim if i else input_dim
            self.convs.append(pyg_nn.Sequential('x, edge_index, edge_attr', [
                (pyg_nn.convs.GATv2Conv(
                    in_dim,
                    hidden_dim,
                    num_heads,
                    concat=False,
                    edge_dim=edge_dim,
                    negative_slope=negative_slope,
                ), 'x, edge_index, edge_attr -> x'),
                pyg_nn.norm.GraphNorm(in_channels=hidden_dim),
                nn.PReLU(num_parameters=hidden_dim),
            ]))

        self.post_mp = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.Dropout(dropout),
            nn.PReLU(num_parameters=hidden_dim),
            nn.Linear(hidden_dim, 1),
            nn.Sigmoid(),
        )

    def forward(self, data: DataBatch) -> Tensor:
        '''
        Apply a forward pass for a data batch.

        :param data: torch_geometric.data.batch.DataBatch instance to process
        :return: torch.Tensor with class probabilities for each data instance in a batch
        '''

        x, edge_index, edge_attr, batch = data.x, data.edge_index, data.edge_attr, data.batch

        for conv in self.convs:
            x = conv(x, edge_index, edge_attr)
        x = pyg_nn.global_max_pool(x, batch)
        x = self.post_mp(x)

        return x


def train_gnn(
    model: nn.Module,
    train_loader: DataLoader,
    n_epochs: int,
    optimizer: Optimizer,
    scheduler: LRScheduler | None = None,
    eval_loss: Callable[[Tensor, Tensor], Tensor] = fn.binary_cross_entropy,
    val_loader: DataLoader | None = None,
    *,
    verbose: bool = True,
) -> tuple[nn.Module, list[float], list[float], list[float]]:
    '''
    Train a graph neural network classifier.

    :param model: model to train
    :param train_loader: data loader for training
    :param n_epochs: number of epochs to train a model for
    :param optimizer: PyTorch Optimizer instance
    :param scheduler: PyTorch LRScheduler instance
    :param eval_loss: callback to evaluate loss function on a data batch
    :param val_loader: data loader for validation
    :param verbose: verbosity flag
    :return: trained model, list of mean batch loss per training epoch, \
        list of train accuracy values, and list of validation accuracy values
    '''

    train_loss = []
    train_accuracy = []
    val_accuracy = []

    device = next(model.parameters()).device
    train_size = len(train_loader.dataset)
    val_size = len(val_loader.dataset) if val_loader is not None else 0

    with tqdm(range(n_epochs), desc='Training GNN...', total=n_epochs, disable=not verbose) as bar:
        for epoch in bar:
            batch_train_loss = []
            batch_train_accuracy = 0
            batch_val_accuracy = 0

            model.train()
            for batch in train_loader:
                batch = batch.to(device)
                logits = model(batch)[:, 0]
                labels = batch.y.type(torch.float)

                loss = eval_loss(logits, labels)

                loss.backward()
                optimizer.step()
                if scheduler is not None:
                    scheduler.step()
                optimizer.zero_grad()

                batch_train_loss.append(float(loss.data.numpy()))
                pred = (logits > 0.5)
                batch_train_accuracy += (labels == pred).numpy().sum()
            train_loss.append(float(np.mean(batch_train_loss)))
            train_accuracy.append(batch_train_accuracy / train_size)

            if val_loader is not None:
                model.eval()
                for batch in val_loader:
                    batch = batch.to(device)
                    logits = model(batch)[:, 0]
                    labels = batch.y.type(torch.float)
                    pred = (logits > 0.5)
                    batch_val_accuracy += (labels == pred).numpy().sum()
                val_accuracy.append(batch_val_accuracy / val_size)

                bar.set_postfix(
                    loss=f'{train_loss[-1]:.4f}',
                    train_acc=f'{train_accuracy[-1] * 100:.4f}',
                    val_acc=f'{val_accuracy[-1] * 100:.4f}',
                )
            else:
                bar.set_postfix(
                    loss=f'{train_loss[-1]:.4f}',
                    train_acc=f'{train_accuracy[-1] * 100:.4f}',
                )

    return model, train_loss, train_accuracy, val_accuracy



'''Node and Tree classes.'''

from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Generator, Sequence

import attr

from ..exception import ArgumentError, ParseError
from ..logger import LoggerMixin
from .parser import split_conllu_into_lines
from .tag import Embedding, SyntaxTag, Tag, get_conllu_children, parse_conllu


__all__ = (
    'BaseContextualEmbeddingModel',
    'BaseEmbeddingModel',
    'BaseMorphTagger',
    'BaseStaticEmbeddingModel',
    'Node',
    'Tree',
)


@attr.s(slots=True, kw_only=True, eq=True, repr=False)
class Node:
    '''Tree node with its tag, link to a parent and a list of children nodes.'''

    tag: Tag = attr.ib()
    parent: Node | None = attr.ib(default=None, eq=False)
    children: list[Node] = attr.ib(factory=list)

    @staticmethod
    def from_conllu(
        syntax_tag: SyntaxTag,
        graph: dict[int, list[SyntaxTag]],
        idx2node: dict[int, Node],
    ) -> Node:
        '''
        Node construction from syntactic data obtained from CoNLL-U.

        :param syntax_tag: syntax tag of node to build
        :param graph: mapping from node index to a list of children's syntax tags
        :param idx2node: mapping from node index to a node, updated with the current node
        :return: new Node instance
        '''

        tag = Tag(syntax_tag=syntax_tag)
        node = Node(tag=tag)
        idx = syntax_tag.idx
        idx2node[idx] = node

        node.parent = idx2node.get(syntax_tag.head)
        node.children = [
            Node.from_conllu(child_tag, graph, idx2node)
            for child_tag in graph[idx]
        ]

        return node

    @property
    def idx(self) -> int:
        '''Node index (in a 1-indexed sentence).'''

        return self.tag.syntax_tag.idx

    @property
    def token(self) -> str:
        '''Initial token (before lemmatization).'''

        return self.tag.token

    @property
    def lemma(self) -> str:
        '''Normal form of token.'''

        return self.tag.lemma

    @property
    def deprel(self) -> str:
        '''Dependency relation name.'''

        return self.tag.syntax_tag.deprel

    @property
    def grammemes(self) -> frozenset[str]:
        '''Set of token grammemes.'''

        return self.tag.grammemes

    @property
    def embedding(self) -> Embedding | None:
        '''Token embedding (if present).'''

        return self.tag.embedding

    @property
    def deg(self) -> int:
        '''Node degree in the tree.'''

        return len(self.children) + (self.parent is not None)

    def __repr__(self) -> str:
        return repr(self.tag)


@attr.s(slots=True, kw_only=True, init=False)
class BaseMorphTagger(LoggerMixin, ABC):
    '''Base class for morphological taggers.'''

    def __init__(self, **kwargs: Any):
        '''Tagger initialization.'''

        LoggerMixin.__init__(self, **kwargs)

    @abstractmethod
    def restore_tag(self, node: Node) -> None:
        '''
        Restore MorphTag info for a given node.

        :param node: Node instance to restore tag for
        :return: None
        '''

        ...


@attr.s(slots=True, kw_only=True, init=False)
class BaseEmbeddingModel(LoggerMixin, ABC):
    '''Base class for word embedding models.'''

    def __init__(self, **kwargs: Any):
        '''Embedding model initialization.'''

        LoggerMixin.__init__(self, **kwargs)

    @abstractmethod
    def restore_embedding(self, node: Node) -> None:
        '''
        Restore static embedding for a given node.

        :param node: Node instance to restore embedding for
        :return: None
        '''

        ...

    @abstractmethod
    def process_sentence(self, tokens: list[str]) -> Sequence[Embedding]:
        '''
        Produce embeddings for a list of tokens.

        :param tokens: list of tokens to produce embeddings for
        :return: sequence of embeddings, one per each token
        '''

        ...

    @abstractmethod
    def process_batch(self, tokens: list[list[str]]) -> Sequence[Sequence[Embedding]]:
        '''
        Produce embeddings for tokens, given a list of sentences split into tokens.

        :param tokens: list of lists of tokens, one list per sentence
        :return: sequence of embedding sequences, one sequence per sentence
        '''

        ...

    def is_static(self) -> bool:
        '''
        Check whether a model generated static embeddings for each token.

        :return: True if model produces static vectors, False otherwise
        '''

        return False

    def is_contextual(self) -> bool:
        '''
        Check whether a model generated contextual (dynamic) embeddings for each token.

        :return: True if model produces contextual vectors, False otherwise
        '''

        return False

    @property
    def emb_size(self) -> int:
        '''Embedding dimensionality.'''

        raise NotImplementedError


@attr.s(slots=True, kw_only=True, init=False)
class BaseStaticEmbeddingModel(BaseEmbeddingModel):
    '''Base class for static word embedding models.'''

    def __init__(self, **kwargs: Any):
        '''Static embedding model initialization.'''

        BaseEmbeddingModel.__init__(self, **kwargs)

    def restore_embedding(self, node: Node) -> None:
        '''
        Restore static embedding for a given node.

        :param node: Node instance to restore embedding for
        :return: None
        '''

        pass

    def process_sentence(self, tokens: list[str]) -> Sequence[Embedding]:
        '''
        Produce embeddings for a list of tokens.

        :param tokens: list of tokens to produce embeddings for
        :return: sequence of embeddings, one per each token
        '''

        raise NotImplementedError

    def process_batch(self, tokens: list[list[str]]) -> Sequence[Sequence[Embedding]]:
        '''
        Produce embeddings for tokens, given a list of sentences split into tokens.

        :param tokens: list of lists of tokens, one list per sentence
        :return: sequence of embedding sequences, one sequence per sentence
        '''

        raise NotImplementedError

    def is_static(self) -> bool:
        '''
        Check whether a model generated static embeddings for each token.

        :return: True if model produces static vectors, False otherwise
        '''

        return True


@attr.s(slots=True, kw_only=True, init=False)
class BaseContextualEmbeddingModel(BaseEmbeddingModel):
    '''Base class for contextual word embedding models.'''

    def __init__(self, **kwargs: Any):
        '''Contextual embedding model initialization.'''

        BaseEmbeddingModel.__init__(self, **kwargs)

    def restore_embedding(self, node: Node) -> None:
        '''
        Restore static embedding for a given node.

        :param node: Node instance to restore embedding for
        :return: None
        '''

        raise NotImplementedError

    def is_contextual(self) -> bool:
        '''
        Check whether a model generated contextual (dynamic) embeddings for each token.

        :return: True if model produces contextual vectors, False otherwise
        '''

        return True


@attr.s(slots=True, kw_only=True, eq=True, repr=False)
class Tree:
    '''Attributed syntactic (dependency) tree.'''

    root: Node = attr.ib()
    sentence: str = attr.ib(eq=False)
    idx2node: dict[int, Node] = attr.ib(factory=dict, eq=False)

    TAB_SIZE = 4

    @staticmethod
    def from_conllu(
        conllu: str,
        morph_tagger: BaseMorphTagger | None = None,
        *,
        static_emb_model: BaseEmbeddingModel | None = None,
        embs: Sequence[Embedding] | None = None,
    ) -> Tree:
        '''
        Tree construction from sentence CoNLL-U data.

        :param morph_tagger: morphological tagger
        :param static_emb_model: static embedding model
        :param embs: list of token embeddings obtained from a contextual model in advance
        :return: new Tree instance
        '''

        if static_emb_model is not None and embs is not None:
            raise ArgumentError(
                'Cannot pass both "static_emb_model" and "embs" '
                'arguments to build Tree.'
            )

        for line in split_conllu_into_lines(conllu):
            if line.startswith('# text = '):
                sentence = line[9:]
                break
        else:
            raise ParseError('Field "text" not found in CoNLL-U data.')

        idx2tag = parse_conllu(conllu)
        graph = get_conllu_children(idx2tag)
        root_tag = graph[0][0]

        idx2node: dict[int, Node] = {}
        root = Node.from_conllu(root_tag, graph, idx2node)
        tree = Tree(sentence=sentence, root=root, idx2node=idx2node)

        if morph_tagger is not None:
            tree.restore_morph_tags(morph_tagger)

        if static_emb_model is not None:
            tree.restore_static_embeddings(static_emb_model)
        elif embs is not None:
            tree.assign_contextual_embeddings(embs)
        else:
            pass

        return tree

    def restore_morph_tags(self, morph_tagger: BaseMorphTagger) -> Tree:
        '''
        Use morphological tagger to restore MorphTag node attributes.

        :param morph_tagger: morphological tagger to use
        :return: self
        '''

        for node in self.idx2node.values():
            morph_tagger.restore_tag(node)
        return self

    def restore_static_embeddings(self, static_emb_model: BaseEmbeddingModel) -> Tree:
        '''
        Use static embedding model to restore word embeddings at each node.

        :param static_emb_model: static embedding model to use
        :return: self
        '''

        for node in self.idx2node.values():
            static_emb_model.restore_embedding(node)
        return self

    def assign_contextual_embeddings(self, embs: Sequence[Embedding]) -> Tree:
        '''
        Assign (probably contextual) word embeddings for each node.

        :param embs: list of token embeddings obtained from a contextual model in advance
        :return: self
        '''

        if len(embs) != self.size:
            raise ArgumentError(
                'Wrong number of embeddings for a sentence: '
                f'expected {self.size}, get {len(embs)}'
            )

        for idx, node in self.idx2node.items():
            node.tag.embedding = embs[idx - 1]
        return self

    def __iter__(self) -> Generator[Node, None, None]:
        yield from self.idx2node.values()

    def __getitem__(self, idx: int) -> Node:
        try:
            return self.idx2node[idx]
        except KeyError:
            raise IndexError(f'No such node with index {idx} in the tree.')

    def __repr__(self) -> str:
        return f'Tree({self.sentence})'

    def pprint(self) -> None:
        '''
        Pretty printing (is it?).

        :return: None
        '''

        traversed = self._traverse(self.root)
        print(f'{self}:\n{traversed}')

    @staticmethod
    def _traverse(node: Node, depth: int = 0) -> str:
        return (
            ' ' * Tree.TAB_SIZE * depth +
            str(node) +
            ('\n' if node.children else '') +
            '\n'.join(
                Tree._traverse(child, depth + 1)
                for child in node.children
            )
        )

    @property
    def size(self) -> int:
        '''Number of vertices in a tree.'''

        return len(self.idx2node)



'''SemSim algo package.'''

from .default import *
from .engine import *
from .filter import *
from .heuristic import *
from .kernel import *
from .neural import *
from .parser import *
from .representation import *
from .tag import *
from .tree import *

__all__ = (
    'AutoEmbeddingModel',
    'CBOWModel',
    'ELMoModel',
    'PymorphyTagger',
    'UDPipeSyntaxParser',

    'SemSimEngine',

    'BaseSentenceFilter',
    'IdentitySentenceFilter',
    'MinLengthSentenceFilter',

    'BaseOrderHeuristic',
    'JaccardIndexOrderHeuristic',
    'PairwiseOrderHeuristic',

    'AdditiveKernel',
    'AssignmentKernel',
    'BaseKernel',
    'CompositeKernel',
    'MSK',
    'NeuralKernel',
    'SABK',
    'TABK',

    'GAT',
    'GCN',
    'build_pyg_dataset',
    'get_edge_ohe',
    'train_gnn',

    'BaseSyntaxParser',
    'clean_text',
    'extract_text',
    'split_conllu_into_lines',
    'split_conllu_into_tokens',
    'split_conllu_sentence_tokens',

    'TextEngine',

    'DEPREL_TO_IDX',
    'Embedding',
    'MorphTag',
    'SyntaxTag',
    'Tag',
    'get_conllu_children',
    'parse_conllu',

    'BaseContextualEmbeddingModel',
    'BaseMorphTagger',
    'BaseStaticEmbeddingModel',
    'Node',
    'Tree',
)



'''Module for tree attributes (tags).'''

from collections import defaultdict

import attr
from attr import validators
from torch import FloatTensor

from .parser import split_conllu_into_lines


__all__ = (
    'DEPREL_TO_IDX',

    'Embedding',
    'MorphTag',
    'SyntaxTag',
    'Tag',

    'get_conllu_children',
    'parse_conllu',
)


DEPREL_TO_IDX = {
    'acl': 0,
    'acl:relcl': 1,
    'advcl': 2,
    'advmod': 3,
    'amod': 4,
    'appos': 5,
    'aux': 6,
    'aux:pass': 7,
    'case': 8,
    'cc': 9,
    'ccomp': 10,
    'compound': 11,
    'conj': 12,
    'cop': 13,
    'csubj': 14,
    'csubj:pass': 15,
    'dep': 16,
    'det': 17,
    'discourse': 18,
    'dislocated': 19,
    'expl': 20,
    'fixed': 21,
    'flat': 22,
    'flat:foreign': 23,
    'flat:name': 24,
    'iobj': 25,
    'list': 26,
    'mark': 27,
    'nmod': 28,
    'nsubj': 29,
    'nsubj:pass': 30,
    'nummod': 31,
    'nummod:entity': 32,
    'nummod:gov': 33,
    'obj': 34,
    'obl': 35,
    'obl:agent': 36,
    'obl:tmod': 37,
    'orphan': 38,
    'parataxis': 39,
    'punct': 40,
    'root': 41,
    'vocative': 42,
    'xcomp': 43,
}


@attr.s(slots=True, kw_only=True, init=False, eq=True, repr=False)
class SyntaxTag:
    '''Tree node tag with syntactic information extracted from CoNLL-U.'''

    idx: int = attr.ib(eq=False, validator=validators.gt(0))
    form: str = attr.ib()
    lemma: str = attr.ib()
    upos: str = attr.ib(eq=False)
    feats: dict[str, str] = attr.ib()
    head: int = attr.ib(eq=False, validator=validators.ge(0))
    deprel: str = attr.ib()
    deps: dict[int, str] = attr.ib(eq=False)
    misc: dict[str, str] = attr.ib(eq=False)

    def __init__(self, conllu_line: str):
        '''
        SyntaxTag initialization.

        :param conllu_line: one line of CoNLL-U data corresponding to the current node
        :return: None
        '''

        idx, form, lemma, upos, _, feats, head, deprel, deps, misc = conllu_line.split('\t')

        self.idx = int(idx)
        self.form = form
        self.lemma = lemma
        self.upos = upos

        self.feats = dict(
            feature.split('=') for feature in feats.split('|')
        ) if feats != '_' else {}

        self.head = int(head)
        self.deprel = deprel

        self.deps = {
            int(head): relation
            for head, relation in map(
                lambda dep: dep.split(':'),
                deps.split('|')
            )
        } if deps != '_' else {}

        self.misc = dict(
            attr.split('=') for attr in misc.split('|')
        ) if misc != '_' else {}

    def __repr__(self) -> str:
        return f'--<{self.deprel}>--> [{self.idx}] {self.form}'


def parse_conllu(conllu: str) -> dict[int, SyntaxTag]:
    '''
    Extract syntactic info from full parsed CoNLL-U data.

    :param conllu: a string containing full CoNLL-U data
    :return: mapping from node index to its syntax tag
    '''

    idx2tag = {}
    for line in split_conllu_into_lines(conllu):
        if line.startswith('#'):
            continue
        tag = SyntaxTag(line)
        idx2tag[tag.idx] = tag
    return idx2tag


def get_conllu_children(idx2tag: dict[int, SyntaxTag]) -> dict[int, list[SyntaxTag]]:
    '''
    Transpose trees to gather syntactic info about children of each node.

    :param idx2tag: mapping from node index to its syntax tag
    :return: mapping from node index to a list of syntax tags of its children
    '''

    graph = defaultdict(list)
    for idx, tag in idx2tag.items():
        graph[tag.head].append(tag)
    return graph


@attr.s(slots=True, kw_only=True)
class MorphTag:
    '''Tree node tag with morpholigical info (such as lemma, POS tag, and set of grammemes).'''

    normal_form: str = attr.ib()
    pos: str = attr.ib(default='NOUN')
    grammemes: frozenset[str] = attr.ib(factory=frozenset)

    @property
    def lemma(self) -> str:
        '''Normal form of token.'''

        return self.normal_form


Embedding = FloatTensor


@attr.s(slots=True, kw_only=True, eq=True, repr=False)
class Tag:
    '''Tree node tag with all the information about the node and its token.'''

    syntax_tag: SyntaxTag = attr.ib()
    morph_tag: MorphTag = attr.ib(default=None)
    embedding: Embedding | None = attr.ib(default=None, eq=False)

    @property
    def token(self) -> str:
        '''Initial token (before lemmatization).'''

        return self.syntax_tag.form

    @property
    def lemma(self) -> str:
        '''Normal form of token.'''

        return self.morph_tag.lemma

    @property
    def grammemes(self) -> frozenset[str]:
        '''Set of token grammemes.'''

        return self.morph_tag.grammemes

    def __repr__(self) -> str:
        return (
            f'{self.syntax_tag}' +
            f' ({self.lemma}: {self.morph_tag.pos})' +
            (' *no-emb!*' if self.embedding is None else '')
        )



'''Parsed text representation.'''

from io import StringIO
from pathlib import Path
from typing import Any, Generator, cast

import attr
from tqdm import tqdm

from .default import PymorphyTagger, UDPipeSyntaxParser
from ..exception import ArgumentError, BaseSemSimException
from .filter import BaseSentenceFilter, MinLengthSentenceFilter
from ..logger import LoggerMixin, get_logger
from .parser import BaseSyntaxParser, clean_text, extract_text, split_conllu_into_tokens
from .tree import BaseEmbeddingModel, BaseMorphTagger, BaseStaticEmbeddingModel, Tree


__all__ = (
    'TextEngine',
    'TextRepr',
)


logger = get_logger(__name__)


@attr.s(slots=True, kw_only=True, init=False, repr=False)
class TextRepr(LoggerMixin):
    '''Text representation as a list of trees (one per sentence).'''

    text: str = attr.ib()
    trees: list[Tree] = attr.ib()

    def __init__(
        self,
        text: str,
        syntax_parser: BaseSyntaxParser,
        morph_tagger: BaseMorphTagger,
        emb_model: BaseEmbeddingModel,
        sentence_filter: BaseSentenceFilter,
        *,
        verbose: bool = True,
    ):
        '''
        Representation initialization using a bunch of models.

        :param text: a text to represent
        :param syntax_parser: syntax model to obtain CoNLL-U data and split the text into sentences and tokens
        :param morph_tagger: morphological tagger to restore MorphTag nodes' data
        :param emb_model: embedding model to use for restoring word embeddings at each node
        :param sentence_filter: sentence filter for throwing trash sentences away (e.g. too short ones)
        :param verbose: verbosity flag
        '''

        LoggerMixin.__init__(self, verbose=verbose, logger=logger)

        text = clean_text(text)
        conllu = syntax_parser.parse_text(text)
        conllu_sentences = list(filter(
            sentence_filter.is_proper,
            conllu.strip().split('\n\n'),
        ))

        if emb_model.is_static():
            self.log('Constructing trees...')
            trees = [
                Tree.from_conllu(
                    sentence,
                    morph_tagger,
                    static_emb_model=cast(BaseStaticEmbeddingModel, emb_model),
                )
                for sentence in tqdm(
                    conllu_sentences,
                    desc='Tree construction',
                    total=len(conllu_sentences),
                    disable=not verbose,
                )
            ]
        elif emb_model.is_contextual():
            tokens = split_conllu_into_tokens(conllu, verbose=verbose)
            all_embs = emb_model.process_batch(tokens)

            self.log('Constructing trees...')
            trees = [
                Tree.from_conllu(sentence, morph_tagger, embs=embs)
                for sentence, embs in tqdm(
                    zip(conllu_sentences, all_embs),
                    desc='Tree construction',
                    total=len(conllu_sentences),
                    disable=not verbose,
                )
            ]
        else:
            raise ArgumentError(
                'Parameter "emb_model" should be an instance of either '
                'BaseStaticEmbeddingModel, BaseContextualEmbeddingModel '
                'or AutoEmbeddingModel class.'
            )

        self.__attrs_init__(verbose=verbose, logger=logger, text=text, trees=trees)
        self.log('Successfully processed text.\n')

    def __getitem__(self, idx: int) -> Tree:
        return self.trees[idx]

    def __iter__(self) -> Generator[Tree, None, None]:
        yield from self.trees

    def __len__(self) -> int:
        return len(self.trees)

    def __repr__(self) -> str:
        return '\n\n'.join(
            f'Sentence #{idx}\n{tree}'
            for idx, tree in enumerate(self.trees)
        )


@attr.s(slots=True, kw_only=True, init=False)
class TextEngine(LoggerMixin):
    '''Composite model for processing texts and representing them as a list of trees.'''

    syntax_parser: BaseSyntaxParser = attr.ib()
    morph_tagger: BaseMorphTagger = attr.ib()
    emb_model: BaseEmbeddingModel = attr.ib()
    sentence_filter: BaseSentenceFilter = attr.ib(factory=MinLengthSentenceFilter)

    def __init__(self, *, verbose: bool = True, **kwargs: Any):
        '''
        Engine initialization.

        :param verbose: verbosity flag
        '''

        self.__attrs_init__(verbose=verbose, logger=logger, **kwargs)
        self.log('Successfully built TextEngine.\n')

    @syntax_parser.default
    def _(self) -> BaseSyntaxParser:
        return UDPipeSyntaxParser(verbose=self.verbose)

    @morph_tagger.default
    def _(self) -> BaseMorphTagger:
        return PymorphyTagger(verbose=self.verbose)

    @emb_model.default
    def _(self) -> BaseEmbeddingModel:
        return BaseStaticEmbeddingModel(verbose=self.verbose)

    def __attrs_post_init__(self) -> None:
        self.set_verbosity(self.verbose)

    def set_verbosity(self, verbose: bool) -> None:
        '''
        Set verbosity value.

        :param verbose: verbosity flag
        :return: None
        '''

        super().set_verbosity(verbose)
        for model in (self.syntax_parser, self.morph_tagger, self.emb_model):
            model.set_verbosity(verbose)

    def _process(self, text: str) -> TextRepr:
        return TextRepr(
            text,
            self.syntax_parser,
            self.morph_tagger,
            self.emb_model,
            self.sentence_filter,
            verbose=self.verbose,
        )

    def process(self, arg: str | StringIO | Path, is_path: bool = False) -> TextRepr:
        '''
        Process a text source and construct its representation.

        :param arg: a string, or buffer, or path to a file to be processed
        :param is_path: True if first argument is a string containing a path to a file
        :return: new TextRepr instance for the processed text
        '''

        if isinstance(arg, StringIO):
            return self._process(arg.read())
        if is_path or isinstance(arg, Path):
            try:
                text = extract_text(arg, verbose=self.verbose)
                return self._process(text)
            except BaseSemSimException:
                raise
        return self._process(arg)



'''SemSim library.'''

from .algo import *
from .download import *
from .exception import *


__all__ = (
    # algo package modules
    'AutoEmbeddingModel',
    'CBOWModel',
    'ELMoModel',
    'PymorphyTagger',
    'UDPipeSyntaxParser',

    'SemSimEngine',

    'BaseSentenceFilter',
    'IdentitySentenceFilter',
    'MinLengthSentenceFilter',

    'BaseOrderHeuristic',
    'JaccardIndexOrderHeuristic',
    'PairwiseOrderHeuristic',

    'AdditiveKernel',
    'AssignmentKernel',
    'BaseKernel',
    'CompositeKernel',
    'MSK',
    'NeuralKernel',
    'SABK',
    'TABK',

    'BaseSyntaxParser',
    'clean_text',
    'extract_text',
    'split_conllu_into_lines',
    'split_conllu_into_tokens',
    'split_conllu_sentence_tokens',

    'TextEngine',

    'DEPREL_TO_IDX',
    'Embedding',
    'MorphTag',
    'SyntaxTag',
    'Tag',
    'get_conllu_children',
    'parse_conllu',

    'BaseContextualEmbeddingModel',
    'BaseMorphTagger',
    'BaseStaticEmbeddingModel',
    'Node',
    'Tree',

    # download.py
    'download',
    'download_all',
    'get_model_path',

    # exception.py
    'ArgumentError',
    'BaseSemSimException',
    'FetchError',
    'ModelError',
    'ParseError',
)



'''Tree comparison kernels implementation.'''

from abc import ABC, abstractmethod
from collections import Counter
from pathlib import Path
from typing import Any, Callable, cast

import attr
from attr import validators
import networkx as nx
from networkx.algorithms.bipartite.matching import minimum_weight_full_matching
import numpy as np
import numpy.typing as npt
from scipy.spatial.distance import (
    chebyshev, cityblock, cosine, euclidean, hamming, jaccard, minkowski,
)
import torch
from torch import nn
from torch_geometric.loader import DataLoader

from ..download import get_model_path
from ..exception import ArgumentError
from .neural import build_pyg_dataset, GAT, GCN
from .tag import DEPREL_TO_IDX, Embedding
from .tree import Node, Tree


__all__ = (
    'AdditiveKernel',
    'AssignmentKernel',
    'BaseKernel',
    'CompositeKernel',
    'MSK',
    'NeuralKernel',
    'SABK',
    'TABK',

    'jaccard_index',
)


DistSimFunction = Callable[[Embedding, Embedding], float]
DistSimMatrix = npt.NDArray[np.float32]

DEFAULT_LAMBDA = 3
DISTANCES: dict[str, DistSimFunction] = {
    'chebyshev': chebyshev,
    'cosine': cosine,
    'euclidean': euclidean,
    'hamming': hamming,
    'jaccard': jaccard,
    'manhattan': cityblock,
}


def get_similarity_func(
    func: str | DistSimFunction,
    is_distance: bool | None = None,
    is_normalized: bool | None = None,
    p: float | None = None,
) -> DistSimFunction:
    '''
    Get similarity metric for token embeddings given user-defined parameters.

    :param func: similarity/distance metric function or name of one of predefined distance metrics
    :param is_distance: True if user-defined callable `func` is distance metric, not similarity metric
    :param is_normalized: True if user-defined callable `func` returns valued from range [0; 1]
    :param p: power parameter for Minkowski distance
    :return: similarity metric function
    '''

    if isinstance(func, str):
        if is_distance is not None or is_normalized is not None:
            raise ArgumentError(
                'You cannot specify "is_distance" and "is_normalized" '
                'parameters while providing a metric name as a string.')
        if func == 'minkowski':
            if p is None:
                raise ArgumentError('Specify value of parameter "p" for Minkowski distance.')
            if p <= 0:
                raise ArgumentError('Parameter "p" should be positive.')
            return lambda x, y: 1 / (1 + minkowski(x, y, p))
        if func not in DISTANCES:
            raise ArgumentError(f'Unknown metric "{func}".')
        if p is not None:
            raise ArgumentError('Cannot specify parameter "p" for non-Minkowski distance.')

        dist = DISTANCES[func]
        if func == 'cosine':
            return lambda x, y: 1 - dist(x, y) / 2
        if func in {'jaccard', 'hamming'}:
            return lambda x, y: 1 - dist(x, y)
        return lambda x, y: 1 / (1 + dist(x, y))

    if not callable(func):
        raise ArgumentError('Parameter "func" should be either a string or callable.')

    if is_distance:
        if is_normalized:
            return cast(DistSimFunction, lambda x, y: 1 - func(x, y))
        return cast(DistSimFunction, lambda x, y: 1 / (1 + func(x, y)))
    if is_normalized is not None:
        raise ArgumentError(
            'Parameter "is_normalized" is not used '
            'for custom similarity functions.')
    return func


def build_deprel_sim_matrix(
    deprel_sim: DistSimMatrix | None = None,
    theta: float | None = None,
    lambda_: float | None = None,
) -> DistSimMatrix:
    r'''
    Construct dependency relation similarity matrix.

    :param deprel_sim: user-defined dependency relation similarity matrix
    :param theta: $\theta$ value placed on the matrix diagonal
    :param lambda_: $\lambda$ value added to the matrix diagonal
    :return: dependency relation similarity matrix
    '''

    num_deps = len(DEPREL_TO_IDX)
    if deprel_sim is not None:
        if theta is not None or lambda_ is not None:
            raise ArgumentError(
                'You should specify either "deprel_sim" or "theta" '
                'or "lambda_" parameter.')
        if deprel_sim.shape != (num_deps, num_deps):
            raise ArgumentError(
                f'Wrong shape {deprel_sim.shape} of "deprel_sim": '
                'expected {(num_deps, num_deps)}.')
        if min(deprel_sim) < 0:
            raise ArgumentError(
                'The "deprel_sim" matrix should only contain '
                'non-negative numbers.')
        return deprel_sim

    if theta is not None:
        if lambda_ is not None:
            raise ArgumentError(
                'You should specify either "theta" or "lambda_"'
                'parameter, but not both.')
        if theta < 1:
            raise ArgumentError('Parameter "theta" cannot be less than 1.')
        lambda_ = theta - 1

    if lambda_ is not None:
        if lambda_ < 0:
            raise ArgumentError('Parameter "lambda_" should be non-negative.')
    else:
        lambda_ = DEFAULT_LAMBDA

    deprel_sim = np.ones((num_deps, num_deps), dtype=float)
    deprel_sim += lambda_ * np.eye(num_deps, dtype=float)
    deprel_sim /= lambda_ + 1
    return deprel_sim


def jaccard_index(lhs: set[str] | frozenset[str], rhs: set[str] | frozenset[str]) -> float:
    '''
    Compute Jaccard similarity index for two given sets of strings.

    :param lhs: left-hand side operand, set of strings
    :param rhs: right-hand side operand, set of strings
    :return: Jaccard index
    '''

    intersection = lhs & rhs
    return len(intersection) / (len(lhs) + len(rhs) - len(intersection))


class BaseKernel(ABC):
    '''Base class for tree comparison kernels.'''

    @abstractmethod
    def compute(self, lhs: Tree, rhs: Tree) -> float:
        '''
        Compute similarity score for two given trees.

        :param lhs: left-hand side operand, attributed tree of a sentence
        :param rhs: right-hand side operand, attributed tree of a sentence
        :return: tree similarity score
        '''

        ...


@attr.s(slots=True, kw_only=True, init=False)
class BigramKernel(BaseKernel):
    '''Base class for bigram tree comparison kernels.'''

    cmp_lemma: bool = attr.ib()
    similarity_type: str = attr.ib()
    emb_similarity: DistSimFunction = attr.ib(repr=False)
    deprel_sim: DistSimMatrix = attr.ib(repr=False)

    def __init__(
        self,
        *,
        cmp_lemma: bool = True,

        emb_metric: str | DistSimFunction = 'cosine',
        is_distance: bool | None = None,
        is_normalized: bool | None = None,
        p: float | None = None,

        deprel_sim: DistSimMatrix | None = None,
        theta: float | None = None,
        lambda_: float | None = None,
    ):
        r'''
        Bigram kernel initialization.

        :param cmp_lemma: whether to compare lemmas instead of tokens in case of missing embeddings
        :param emb_metric: custom callable or name of a predefined distance metric for embeddings
        :param is_distance: True if user-defined callable `func` is distance metric, not similarity metric
        :param is_normalized: True if user-defined callable `func` returns valued from range [0; 1]
        :param p: power parameter for Minkowski distance
        :param deprel_sim: user-defined dependency relation similarity matrix
        :param theta: $\theta$ value placed on the matrix diagonal
        :param lambda_: $\lambda$ value added to the matrix diagonal
        '''

        BigramKernel.__attrs_init__(
            self,
            cmp_lemma=cmp_lemma,
            similarity_type=emb_metric if isinstance(emb_metric, str) else 'custom',
            emb_similarity=get_similarity_func(emb_metric, is_distance, is_normalized, p),
            deprel_sim=build_deprel_sim_matrix(deprel_sim, theta, lambda_),
        )


@attr.s(slots=True, kw_only=True, init=False)
class AdditiveKernel(BigramKernel, ABC):  # type: ignore[override]
    '''Base class for additive bigram kernels.'''

    gamma: float = attr.ib(validator=validators.and_(validators.gt(0), validators.le(1)))

    def __init__(self, gamma: float = 0.15, **kwargs: Any):
        r'''
        Additive kernel initialization.

        :param gamma: $\gamma$ value for post-calibrating similarity scores
        '''

        if not 0 < gamma <= 1:
            raise AttributeError('Parameter "gamma" should be in range (0; 1].')
        BigramKernel.__init__(self, **kwargs)
        self.gamma = gamma

    def compute(self, lhs: Tree, rhs: Tree) -> float:
        '''
        Compute similarity score for two given trees.

        :param lhs: left-hand side operand, attributed tree of a sentence
        :param rhs: right-hand side operand, attributed tree of a sentence
        :return: tree similarity score
        '''

        if (lhs.size == 1) ^ (rhs.size == 1):
            return 0
        if lhs == rhs:
            return 1

        self._precompute(lhs, rhs)
        sim = 0.
        for dep_left in lhs:
            if dep_left.parent is None:
                continue
            for dep_right in rhs:
                if dep_right.parent is not None:
                    sim += self._sim(dep_left, dep_right)
        denom = self._denominator(lhs, rhs)
        result = sim / denom if denom else 0
        return result ** self.gamma

    def _precompute(self, lhs: Tree, rhs: Tree) -> None:
        pass

    def _denominator(self, lhs: Tree, rhs: Tree) -> float:
        return max(1, (lhs.size - 1) * (rhs.size - 1))

    def _sim(self, lhs: Node, rhs: Node) -> float:
        sim = self._bigram_weight(lhs, rhs) * self._node_sim(lhs, rhs)
        sim += self._bigram_weight(lhs.parent, rhs.parent) * self._node_sim(lhs.parent, rhs.parent)  # type: ignore
        sim *= self._deprel_sim(lhs.deprel, rhs.deprel) / 2
        return sim

    def _bigram_weight(self, lhs: Node, rhs: Node) -> float:
        return self._node_weight(lhs) * self._node_weight(rhs)

    def _node_weight(self, node: Node, **kwargs: Any) -> float:
        return 1

    def _node_sim(self, lhs: Node, rhs: Node) -> float:
        if lhs.embedding is not None and rhs.embedding is not None:
            return self.emb_similarity(lhs.embedding, rhs.embedding)
        elif self.cmp_lemma:
            return jaccard_index(lhs.grammemes, rhs.grammemes) if lhs.lemma == rhs.lemma else 0
        else:
            return float(lhs.token.lower() == rhs.token.lower())

    def _deprel_sim(self, lhs: str, rhs: str) -> float:
        row = DEPREL_TO_IDX[lhs]
        col = DEPREL_TO_IDX[rhs]
        return self.deprel_sim[row, col]


@attr.s(slots=True, kw_only=True, init=False)
class SABK(AdditiveKernel):  # type: ignore[override]
    '''Simple approximate bigram kernel implementation.'''

    def __init__(self, **kwargs: Any):
        '''Simple approximate bigram kernel initialization.'''

        AdditiveKernel.__init__(self, **kwargs)


@attr.s(slots=True, kw_only=True, init=False)
class TABK(AdditiveKernel):  # type: ignore[override]
    '''TF-IDF based approximate bigram kernel implementation.'''

    idf: dict[str, float] = attr.ib(repr=False)
    use_lemma: bool = attr.ib()
    unknown_idf: float = attr.ib(validator=validators.ge(0))

    _lhs_tf: dict[str, int] = attr.ib(factory=dict, init=False, repr=False)
    _rhs_tf: dict[str, int] = attr.ib(factory=dict, init=False, repr=False)

    def __init__(
        self,
        *,
        idf: dict[str, float] | None = None,
        use_lemma: bool = True,
        unknown_idf: float | None = None,
        **kwargs: Any,
    ):
        '''
        TF-IDF based approximate bigram kernel initialization.

        :param idf: IDF dictionary
        :param use_lemma: whether to use lemmas instead of tokens for lookups in IDF dictionary
        :param unknown_idf: default IDF values for unknown dictionary entries
        '''

        AdditiveKernel.__init__(self, **kwargs)

        self.idf = idf or {}
        self.use_lemma = use_lemma
        if unknown_idf is None:
            unknown_idf = 1 + np.log(len(idf)) if idf else 1
        self.unknown_idf = unknown_idf

    def _precompute(self, lhs: Tree, rhs: Tree) -> None:
        self._lhs_tf = self._compute_tf(lhs)
        self._rhs_tf = self._compute_tf(rhs)

    def _compute_tf(self, tree: Tree) -> dict[str, int]:
        if self.use_lemma:
            return Counter(node.lemma for node in tree)
        return Counter(node.token.lower() for node in tree)

    def _bigram_weight(self, lhs: Node, rhs: Node) -> float:
        return self._node_weight(lhs, tf=self._lhs_tf) * self._node_weight(rhs, tf=self._rhs_tf)

    def _node_weight(self, node: Node, **kwargs: Any) -> float:
        tf: dict[str, int] = kwargs['tf']
        word = node.lemma if self.use_lemma else node.token.lower()
        return tf[word] * self.idf.get(word, self.unknown_idf)

    def _denominator(self, lhs: Tree, rhs: Tree) -> float:
        lhs_s, lhs_s_hat = self._sum_tfidf(lhs.root, tf=self._lhs_tf)
        rhs_s, rhs_s_hat = self._sum_tfidf(rhs.root, tf=self._rhs_tf)

        lhs_s -= self._node_weight(lhs.root, tf=self._lhs_tf)
        rhs_s -= self._node_weight(rhs.root, tf=self._rhs_tf)

        return 0.5 * (lhs_s * rhs_s + (lhs_s_hat - lhs_s) * (rhs_s_hat - rhs_s))

    def _sum_tfidf(self, node: Node, *, tf: dict[str, int]) -> tuple[float, float]:
        s = self._node_weight(node, tf=tf)
        s_hat = s * node.deg

        for child in node.children:
            s_child, s_hat_child = self._sum_tfidf(child, tf=tf)
            s += s_child
            s_hat += s_hat_child

        return s, s_hat


def resolve_alpha_nu(alpha: float | None, nu: float | None) -> tuple[float, float]:
    if alpha is None:
        if nu is None:
            alpha = 0.25
            nu = 1 - alpha
        elif 0 <= nu < 1:
            alpha = 1 - nu
        else:
            raise ArgumentError('Parameter "nu" should be in range [0; 1).')
    elif nu is not None:
        raise ArgumentError(
            'You should specify either "alpha" or "nu" '
            'parameter, but not both.')
    elif 0 < alpha <= 1:
        nu = 1 - alpha
    else:
        raise ArgumentError('Parameter "alpha" should be in range (0; 1].')

    return alpha, nu


@attr.s(slots=True, kw_only=True, init=False)
class MSK(AdditiveKernel):  # type: ignore[override]
    '''Matching subtree kernel implementation.'''

    alpha: float = attr.ib(validator=validators.and_(validators.ge(0), validators.lt(1)))
    nu: float = attr.ib(validator=validators.and_(validators.gt(0), validators.le(1)))

    _dp: list[list[float | None]] = attr.ib(factory=list, init=False, repr=False)

    def __init__(
        self,
        *,
        alpha: float | None = None,
        nu: float | None = None,
        **kwargs: Any,
    ):
        r'''
        Matching subtree kernel initialization.

        :param alpha: $\alpha$ parameter value, fraction of node similarity
        :param nu: $\nu$ parameter value, fraction of subtree similarity
        '''

        self.alpha, self.nu = resolve_alpha_nu(alpha, nu)
        AdditiveKernel.__init__(self, **kwargs)

    def _precompute(self, lhs: Tree, rhs: Tree) -> None:
        self._dp = [
            [None] * rhs.size for _ in range(lhs.size)
        ]

    def _sim(self, lhs: Node, rhs: Node) -> float:
        sim = self._node_sim(lhs.parent, rhs.parent)  # type: ignore
        sim *= self._deprel_sim(lhs.deprel, rhs.deprel)
        sim *= self._subtree_sim(lhs, rhs)
        return sim

    def _subtree_sim(self, lhs: Node, rhs: Node) -> float:
        lhs_idx = lhs.idx - 1
        rhs_idx = rhs.idx - 1
        sim = self._dp[lhs_idx][rhs_idx]
        if sim is not None:
            return sim

        sim = self.alpha * self._node_sim(lhs, rhs)
        if lhs.children and rhs.children:
            children_sim = 0.
            for left in lhs.children:
                for right in rhs.children:
                    dsim = self._subtree_sim(left, right)
                    children_sim += dsim * self._deprel_sim(left.deprel, right.deprel)
            children_sim /= len(lhs.children) * len(rhs.children)
            sim += self.nu * children_sim

        self._dp[lhs_idx][rhs_idx] = sim
        return sim


def resolve_beta_delta(beta: float | None, delta: float | None) -> tuple[float, float]:
    if beta is None:
        if delta is None:
            beta = 0.6
            delta = 1 - beta
        elif 0 <= delta < 1:
            beta = 1 - delta
        else:
            raise ArgumentError('Parameter "delta" should be in range [0; 1].')
    elif delta is not None:
        raise ArgumentError(
            'You should specify either "beta" or "delta" '
            'parameter, but not both.')
    elif 0 <= beta <= 1:
        delta = 1 - beta
    else:
        raise ArgumentError('Parameter "beta" should be in range [0; 1].')

    return beta, delta


@attr.s(slots=True, kw_only=True, init=False)
class CompositeKernel(BigramKernel):  # type: ignore[override]
    '''Composite kernel implementation.'''

    beta: float = attr.ib(validator=validators.and_(validators.ge(0), validators.le(1)))
    delta: float = attr.ib(validator=validators.and_(validators.ge(0), validators.le(1)))
    tabk: TABK = attr.ib()
    msk: MSK = attr.ib()

    def __init__(
        self,
        *,
        gamma: float = 0.15,

        idf: dict[str, float] | None = None,
        use_lemma: bool = True,
        unknown_idf: float | None = None,

        alpha: float | None = None,
        nu: float | None = None,

        beta: float | None = None,
        delta: float | None = None,
        **kwargs: Any,
    ):
        r'''
        Composite kernel initialization.

        :param gamma: $\gamma$ value for post-calibrating similarity scores
        :param idf: IDF dictionary
        :param use_lemma: whether to use lemmas instead of tokens for lookups in IDF dictionary
        :param unknown_idf: default IDF values for unknown dictionary entries
        :param alpha: $\alpha$ parameter value, fraction of node similarity
        :param nu: $\nu$ parameter value, fraction of subtree similarity
        :param beta: $\beta$ parameter value, fraction of TABK score
        :param delta: $\delta$ parameter value, fraction of MSK score
        '''

        BigramKernel.__init__(self, **kwargs)
        self.beta, self.delta = resolve_beta_delta(beta, delta)
        self.tabk = TABK(gamma=gamma, idf=idf or {}, use_lemma=use_lemma, unknown_idf=unknown_idf, **kwargs)
        self.msk = MSK(gamma=gamma, alpha=alpha, nu=nu, **kwargs)

    def compute(self, lhs: Tree, rhs: Tree) -> float:
        '''
        Compute similarity score for two given trees.

        :param lhs: left-hand side operand, attributed tree of a sentence
        :param rhs: right-hand side operand, attributed tree of a sentence
        :return: tree similarity score
        '''

        return self.beta * self.tabk.compute(lhs, rhs) + self.delta * self.msk.compute(lhs, rhs)


@attr.s(slots=True, kw_only=True, init=False)
class AssignmentKernel(BigramKernel):
    '''Assignment kernel based on maximum weight assignment problem.'''

    _dp: list[list[float | None]] = attr.ib(factory=list, init=False, repr=False)

    def __init__(self, **kwargs: Any):
        '''Assignment kernel initialization.'''

        BigramKernel.__init__(self, **kwargs)

    def compute(self, lhs: Tree, rhs: Tree) -> float:
        '''
        Compute similarity score for two given trees.

        :param lhs: left-hand side operand, attributed tree of a sentence
        :param rhs: right-hand side operand, attributed tree of a sentence
        :return: tree similarity score
        '''

        self._dp = [
            [None] * rhs.size for _ in range(lhs.size)
        ]
        score = self._subtree_sim(lhs.root, rhs.root)
        return score

    def _subtree_sim(self, lhs: Node, rhs: Node) -> float:
        lhs_idx = lhs.idx - 1
        rhs_idx = rhs.idx - 1
        sim = self._dp[lhs_idx][rhs_idx]
        if sim is not None:
            return sim

        if not lhs.children or not rhs.children:
            sim = self._node_sim(lhs, rhs)
        else:
            sim = self._max_assignment(lhs.children, rhs.children)
            sim /= len(lhs.children) + len(rhs.children)
            sim *= self._node_sim(lhs, rhs)

        self._dp[lhs_idx][rhs_idx] = sim
        return sim

    def _node_sim(self, lhs: Node, rhs: Node) -> float:
        if lhs.embedding is not None and rhs.embedding is not None:
            return self.emb_similarity(lhs.embedding, rhs.embedding)
        elif self.cmp_lemma:
            return jaccard_index(lhs.grammemes, rhs.grammemes) if lhs.lemma == rhs.lemma else 0
        else:
            return float(lhs.token.lower() == rhs.token.lower())

    def _deprel_sim(self, lhs: str, rhs: str) -> float:
        row = DEPREL_TO_IDX[lhs]
        col = DEPREL_TO_IDX[rhs]
        return self.deprel_sim[row, col]

    def _max_assignment(self, left_nodes: list[Node], right_nodes: list[Node]) -> float:
        top_nodes = [node.idx for node in left_nodes]
        shift = max(top_nodes)

        graph = nx.Graph()
        for lhs in left_nodes:
            for rhs in right_nodes:
                child_sim = self._subtree_sim(lhs, rhs)
                child_sim *= self._deprel_sim(lhs.deprel, rhs.deprel)
                graph.add_edge(lhs.idx, shift + rhs.idx, weight=-child_sim)

        matches = minimum_weight_full_matching(graph, top_nodes=top_nodes, weight='weight')
        score = 0
        for left in top_nodes:
            right = matches.get(left)
            if right is not None:
                score += -2 * graph[left][right]['weight']
        return score


DEFAULT_BATCH_SIZE = 64


@attr.s(slots=True, kw_only=True, init=False)
class NeuralKernel(BaseKernel):
    '''Neural kernel based on graph neural networks.'''

    model: nn.Module = attr.ib()
    device: torch.device = attr.ib(converter=torch.device)
    batch_size: int = attr.ib(validator=validators.gt(0))

    @device.default
    def _(self) -> torch.device:
        return torch.device('cpu')

    def __init__(
        self,
        *,
        model: nn.Module | None = None,
        model_path: str | Path | None = None,
        model_type: str | None = None,
        device: torch.device | str = 'cpu',
        batch_size: int = DEFAULT_BATCH_SIZE,
        **kwargs: Any,
    ):
        '''
        Neural kernel initialization.

        :param model: graph neural network model (torch_geometric)
        :param model_path: path to GNN model checkpoint
        :param model_type: type of GNN to instantiate ('gcn' or 'gat')
        :param device: device to run GNN on ('cpu', 'gpu'/'cuda', 'mps')
        :param batch_size: batch size for batched processing mode
        :param kwargs: other arguments to initialize PyTorch GNN model
        '''

        if model is not None:
            if model_path is not None or model_type is not None:
                raise ArgumentError(
                    'You cannot specify both graph NN model '
                    'and one of the parameters "model_path" and "model_type".'
                )
            device = next(model.parameters()).device
        elif model_type is not None:
            if model_type == 'gcn':
                model = GCN(**kwargs)
            elif model_type == 'gat':
                model = GAT(**kwargs)
            else:
                raise ArgumentError(f'Unknown GNN model type "{model_type}".')
            path = model_path or get_model_path(model_type)
            model.load_state_dict(torch.load(path))
        elif model_path is not None:
            raise ArgumentError(
                'You should specify "model_type" parameter to use "model_path".'
            )
        else:
            model = GCN(**kwargs)
            path = get_model_path('gcn')
            model.load_state_dict(torch.load(path))

        model.eval()
        self.__attrs_init__(
            model=model,
            device=device,
            batch_size=batch_size,
        )

    def compute_batch(self, tree_pairs: list[tuple[Tree, Tree]]) -> list[float]:
        '''
        Compute similarity score for given tree pairs.

        :param tree_pairs: list of pairs of attributed sentence trees
        :return: list of tree similarity score values
        '''

        self.model.eval()
        dataset = build_pyg_dataset(tree_pairs)
        loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=False)
        predictions = []
        for batch in loader:
            y_pred = self.model(batch.to(self.device)).detach()
            predictions.extend(y_pred[:, 0])
        return predictions

    def compute(self, lhs: Tree, rhs: Tree) -> float:
        '''
        Compute similarity score for two given trees.

        :param lhs: left-hand side operand, attributed tree of a sentence
        :param rhs: right-hand side operand, attributed tree of a sentence
        :return: tree similarity score
        '''

        return self.compute_batch([(lhs, rhs)])[0]



'''Fetching standard (built-in) models.'''

from itertools import repeat
from multiprocessing import cpu_count, Pool
import os
from pathlib import Path
from urllib.parse import urlencode

import requests
from tqdm import tqdm

from .exception import FetchError
from .logger import get_logger


__all__ = (
    'download',
    'download_all',
    'get_model_path',
)


logger = get_logger(__name__)


CURDIR = Path(os.path.relpath(__file__)).parent
MODELS_DIR = CURDIR / 'models'

BASE_URL = 'https://cloud-api.yandex.net/v1/disk/public/resources/download?'
FILE_NAMES = {
    'udpipe': 'russian-syntagrus-ud-2.5-191206.udpipe',
    'cbow': 'ruwikiruscorpora_upos_cbow_300_2021.txt',
    'elmo1024': 'ruwikiruscorpora_lemmas_elmo_1024_2019.zip',
    'elmo2048': 'araneum_lemmas_elmo_2048_2020.zip',
    'gcn': 'gcn_model.pth',
    'gat': 'gat_model.pth',
}
FILE_URLS = {
    'udpipe': 'https://disk.yandex.ru/d/Pz17-vtNeasApg',
    'cbow': 'https://disk.yandex.ru/d/6dXzkn0r9Otjnw',
    'elmo1024': 'https://disk.yandex.ru/d/mIZRldOR6B16rw',
    'elmo2048': 'https://disk.yandex.ru/d/fAl1ON3JKmWwDQ',
    'gcn': 'https://disk.yandex.ru/d/kI010Uqj2F51zw',
    'gat': 'https://disk.yandex.ru/d/9_0I88hnEz2G0w',
}
CHUNK_SIZE = 8_192


def get_model_path(file: str) -> Path:
    '''
    Get path to a built-in model if present.

    :param file: model name to find a file for
    :return: path to a model
    '''

    filename = FILE_NAMES.get(file)
    if filename is None:
        raise FetchError(f'Unknown filename "{file}".')
    return MODELS_DIR / filename


def fetch_url(file: str) -> requests.Response:
    '''
    URL fetching worker.

    :param file: filename to fetch
    :return: requests.Response instance
    '''

    file_url = FILE_URLS.get(file)
    if file_url is None:
        raise FetchError(f'Cannot fetch url for file "{file}".')

    api_url = BASE_URL + urlencode(dict(public_key=file_url))
    response = requests.get(api_url)
    download_url = response.json()['href']

    download_response = requests.get(download_url, stream=True)
    if not download_response:
        raise FetchError(
            f'Failed to fetch url for file "{file}": '
            f'HTTP status code {download_response.status_code}'
        )
    return download_response


def download_all(*, verbose: bool = True, force: bool = False) -> None:
    '''
    Download all built-in models in parallel.

    :param verbose: verbosity flag
    :param force: whether to force downloading existing files
    :return: None
    '''

    num_workers = min(cpu_count(), len(FILE_NAMES))
    with Pool(num_workers) as pool:
        pool.starmap(download, zip(FILE_NAMES, repeat(verbose), repeat(force)))


def download(file: str, *, verbose: bool = True, force: bool = False) -> None:
    '''
    Download a built-in model.

    :param file: model name to download
    :param verbose: verbosity flag
    :param force: whether to force downloading existing files
    :return: None
    '''

    if file == 'elmo':
        download('elmo1024')
        download('elmo2048')
        return
    if file == 'neural':
        download('gcn')
        download('gat')
        return

    MODELS_DIR.mkdir(exist_ok=True)
    path = get_model_path(file)
    if not path.exists() or force:
        if verbose:
            logger.info(f'Downloading "{file}"...')

        response = fetch_url(file)
        total = response.headers.get('content-length')
        content = response.iter_content(chunk_size=CHUNK_SIZE)
        bar = tqdm(
            desc=f'Downloading "{file}"...',
            total=int(total) if total is not None else None,
            disable=not verbose,
        )

        with open(path, 'wb') as fout:
            for chunk in content:
                fout.write(chunk)
                bar.update(len(chunk))
        bar.close()
    elif verbose:
        logger.info(f'File "{file}" already exists. Set force=True to proceed anyway.')
    else:
        pass



'''Syntax parsing, CoNLL-U processing and text extraction.'''

from abc import ABC, abstractmethod
from pathlib import Path
import re
from typing import Any, Generator

import textract
from tqdm import tqdm

from ..exception import ParseError
from ..logger import LoggerMixin, get_logger


__all__ = (
    'BaseSyntaxParser',

    'clean_text',
    'extract_text',
    'split_conllu_into_lines',
    'split_conllu_into_tokens',
    'split_conllu_sentence_tokens',
)


logger = get_logger(__name__)


BAD_NON_PRINTABLE_CHARACTERS = ''.join(
    chr(idx)
    for idx in range(1, ord(' '))
    if idx not in (ord(c) for c in '\t\n')
)


def clean_text(text: str) -> str:
    '''
    Remove strange characters from text, usually produced while extracting text from pdf via OCR.

    :param text: source text to clean
    :return: clear text
    '''

    text = re.sub('[\r\n\v\f\x1c\x1d\x1e\x85\u2028\u2029]{2,}', '\n', text)
    text = re.sub(f'[{BAD_NON_PRINTABLE_CHARACTERS}]', ' ', text)
    text = re.sub('[ ]{2,}', ' ', text).strip()
    return text


def extract_text(filename: str | Path, *, verbose: bool = True) -> str:
    '''
    Extract text from a file using OCR when necessary.

    :param filename: path to a file to process
    :param verbose: verbosity flag
    :return: extracted text
    '''

    if verbose:
        logger.info(f'Extracting text from {filename}...')

    try:
        text = textract.process(filename, language='rus').decode()
        return clean_text(text)
    except Exception as exc:  # noqa
        raise ParseError(f'Text extraction failed: {exc}')


class BaseSyntaxParser(LoggerMixin, ABC):
    '''Base class for syntax parsers.'''

    def __init__(self, **kwargs: Any):
        '''Base syntax parses initialization.'''

        LoggerMixin.__init__(self, **kwargs)

    @abstractmethod
    def parse_text(self, text: str) -> str:
        '''
        Parse text and produce its CoNLL-U representation.

        :param text: text to parse
        :return: CoNLL-U representation
        '''

        ...


def split_conllu_into_lines(sentence: str) -> Generator[str, None, None]:
    '''
    Split CoNLL-U sentence data into separate lines (one per token), ignoring comments and empty lines.

    :param sentence: CoNLL-U data for a sentence
    '''

    last_line = ''
    idx = 1
    for line in sentence.strip().splitlines():
        line = line.strip()
        if not line:
            continue

        if line.startswith('#') or line.startswith(f'{idx}\t'):
            if last_line:
                yield last_line
            last_line = line
            idx += line[0].isdigit()
        else:
            last_line += f' {line}'

    if last_line:
        yield last_line


def split_conllu_sentence_tokens(sentence: str) -> list[str]:
    '''
    Get list of tokens from CoNLL-U sentence data.

    :param sentence: CoNLL-U data for a sentence
    :return: list of tokens
    '''

    tokens = []
    for line in split_conllu_into_lines(sentence):
        if not line.startswith('#'):
            _, _, token, *_ = line.split('\t')
            tokens.append(token)
    return tokens


def split_conllu_into_tokens(conllu: str, *, verbose: bool = True) -> list[list[str]]:
    '''
    Split full CoNLL-U data into tokens.

    :param conllu: full parsed CoNLL-U data
    :param verbose: verbosity flag
    :return: list of lists of tokens, one list per each sentence
    '''

    if verbose:
        logger.info('Tokenizing CoNLL-U sentences...')

    try:
        all_tokens = []
        sentences = conllu.strip().split('\n\n')
        for sentence in tqdm(sentences, desc='Sentence tokenization', total=len(sentences), disable=not verbose):
            tokens = split_conllu_sentence_tokens(sentence)
            all_tokens.append(tokens)
        return all_tokens
    except Exception as exc:
        raise ParseError(f'Sentence tokenization failed: {exc}')



'''Standard (built-in) models implementation.'''

from pathlib import Path
from typing import Any, Sequence, cast
from warnings import catch_warnings, simplefilter

import attr
import pymorphy2
from simple_elmo import ElmoModel
import torch
from torch import FloatTensor
from torchwordemb import load_word2vec_text
import ufal.udpipe

from ..download import get_model_path
from ..exception import ArgumentError, FetchError, ModelError
from ..logger import get_logger
from .parser import BaseSyntaxParser
from .tag import Embedding, MorphTag
from .tree import BaseContextualEmbeddingModel, BaseEmbeddingModel, BaseMorphTagger, BaseStaticEmbeddingModel, Node


__all__ = (
    'AutoEmbeddingModel',
    'CBOWModel',
    'ELMoModel',
    'PymorphyTagger',
    'UDPipeSyntaxParser',
)


logger = get_logger(__name__)


PymorphyModel = pymorphy2.analyzer.MorphAnalyzer
UDPipeModel = ufal.udpipe.Model
UDPipeline = ufal.udpipe.Pipeline


@attr.s(slots=True, kw_only=True, init=False)
class UDPipeSyntaxParser(BaseSyntaxParser):
    '''Syntax parser model based on UDPipe.'''

    model_path: str = attr.ib(converter=str)
    model: UDPipeModel = attr.ib(repr=False)    # type: ignore
    pipeline: UDPipeline = attr.ib(repr=False)  # type: ignore

    def __init__(self, model_path: str | Path | None = None, *, verbose: bool = True):
        '''
        Parser initialization.

        :param model_path: path to a UDPipe model to load
        :param verbose: verbosity flag
        '''

        if model_path is None:
            try:
                model_path = get_model_path('udpipe')
            except FetchError:
                raise ModelError(
                    'Cannot load UDPipe model. Consider downloading '
                    'required files using "semsim.download" '
                    'or "semsim.download_all".'
                )

        model = UDPipeModel.load(str(model_path))
        if model is None:
            raise ModelError(
                'Cannot load UDPipe model. Consider downloading '
                'required files using "semsim.download" '
                'or "semsim.download_all".'
            )

        pipeline = UDPipeline(
            model,
            'tokenize',
            ufal.udpipe.Pipeline_DEFAULT_get(),
            ufal.udpipe.Pipeline_DEFAULT_get(),
            'conllu',
        )

        self.__attrs_init__(
            verbose=verbose,
            logger=logger,
            model_path=model_path,
            model=model,
            pipeline=pipeline,
        )
        self.log('Loaded UDPipe model.')

    def parse_text(self, text: str) -> str:
        '''
        Parse text and produce its CoNLL-U representation.

        :param text: text to parse
        :return: CoNLL-U representation
        '''

        self.log('Generating CoNLL-U...')
        return self.pipeline.process(text)  # type: ignore


@attr.s(slots=True, kw_only=True, init=False)
class PymorphyTagger(BaseMorphTagger):
    '''Morphological tagger based on pymorphy2.'''

    model: PymorphyModel = attr.ib(factory=PymorphyModel, init=False, repr=False)  # type: ignore

    __upos2pymorphy: dict[str, set[str]] = {
        'ADJ': {'ADJF', 'ADJS', 'COMP', 'PRTF', 'PRTS'},
        'ADP': {'PREP'},
        'ADV': {'ADVB', 'COMP'},
        'AUX': {'VERB', 'INFN'},
        'CCONJ': {'CONJ'},
        'DET': set(),
        'INTJ': {'INTJ'},
        'NOUN': {'NOUN'},
        'NUM': {'NUMR', 'NUMB', 'ROMN'},
        'PART': {'PRCL'},
        'PRON': {'PRON', 'NRPO'},
        'PROPN': {'NOUN'},
        'PUNCT': {'PNCT'},
        'SCONJ': {'CONJ'},
        'SYM': {'LATN', 'UNKN', 'PNCT', 'ROMN'},
        'VERB': {'VERB', 'INFN', 'GRND', 'PRTF', 'PRTS'},
        'X': {'LATN', 'UNKN', 'PNCT', 'ROMN'},
    }

    def __init__(self, *, verbose: bool = True):
        '''
        PymorphyTagger initialization.

        :param verbose: verbosity flag
        '''

        self.__attrs_init__(verbose=verbose, logger=logger)
        self.log('Loaded pymorphy2 model.')

    def restore_tag(self, node: Node) -> None:
        '''
        Restore MorphTag info for a given node.

        :param node: Node instance to restore tag for
        :return: None
        '''

        syntax_tag = node.tag.syntax_tag
        upos = syntax_tag.upos
        expected_tags = self.__upos2pymorphy.get(upos, set())

        parses = self.model.parse(node.token)  # type: ignore
        picked_parse = parses[0]
        for parse in parses:
            pos = parse.tag.POS
            if pos is not None and ',' in pos:
                pos = pos[:pos.index(',')]
            if pos in expected_tags:
                picked_parse = parse
                break

        node.tag.morph_tag = MorphTag(
            normal_form=picked_parse.normal_form,
            pos=picked_parse.tag.POS,
            grammemes=picked_parse.tag.grammemes,
        )


@attr.s(slots=True, kw_only=True, init=False)
class CBOWModel(BaseStaticEmbeddingModel):  # type: ignore
    '''CBOW embedding model.'''

    path: str = attr.ib(converter=str)
    default: FloatTensor | None = attr.ib(default=None, repr=False)
    vocab: dict[str, int] = attr.ib(repr=False)
    vectors: FloatTensor = attr.ib(repr=False)

    def __init__(
        self,
        path: str | Path | None = None,
        *,
        default: FloatTensor | str | None = 'mean',
        verbose: bool = True,
    ):
        '''
        CBOW embedding model initialization.

        :param path: path to CBOW dictionary and vectors
        :param default: default tensor or a strategy to produce one for missing tokens
        :param verbose: verbosity flag
        '''

        try:
            if path is None:
                path = get_model_path('cbow')
                if not path.exists():
                    raise OSError(f'Cannot load CBOW (word2vec) data from {path}')

            vocab: dict[str, int]
            vectors: FloatTensor
            vocab, vectors = load_word2vec_text(str(path))
            if not vocab:
                raise OSError(f'Cannot load CBOW (word2vec) data from {path}')

            if isinstance(default, str):
                if default == 'mean':
                    default = cast(FloatTensor, torch.mean(vectors, dim=0, dtype=torch.float))
                elif default == 'zero':
                    default = cast(FloatTensor, torch.zeros_like(vectors[0], dtype=torch.float))
                else:
                    idx = vocab.get(default)
                    if idx is None:
                        raise ArgumentError(
                            'Parameter "default" allows only the following string values: '
                            '"mean", "zero", or one of the existing vocabulary tokens.'
                        )
                    default = cast(FloatTensor, vectors[idx])

            if default is not None:
                expected_shape = vectors[0].shape
                real_shape = default.shape
                if real_shape != expected_shape:
                    raise ArgumentError(
                        'Wrong default vector shape: '
                        f'expected {expected_shape}, got {real_shape}'
                    )

            self.__attrs_init__(
                verbose=verbose,
                logger=logger,
                path=path,
                default=default,
                vocab=vocab,
                vectors=vectors,
            )
            self.log('Loaded CBOW model.')
        except (FetchError, OSError):
            raise ModelError(
                'Cannot load CBOW model. Consider downloading '
                'required files using "semsim.download" '
                'or "semsim.download_all".'
            )

    def _lookup(self, token: str) -> Embedding | None:
        idx = self.vocab.get(token)
        if idx is not None:
            return cast(Embedding, self.vectors[idx])
        if self.default is not None:
            return self.default
        return None

    def restore_embedding(self, node: Node) -> None:
        '''
        Restore static embedding for a given node.

        :param node: Node instance to restore embedding for
        :return: None
        '''

        token = f'{node.lemma}_{node.tag.syntax_tag.upos}'
        node.tag.embedding = self._lookup(token)

    @property
    def emb_size(self) -> int:
        '''Embedding dimensionality.'''

        return len(self.vectors[0])


@attr.s(slots=True, kw_only=True, init=False)
class ELMoModel(BaseContextualEmbeddingModel):  # type: ignore
    '''Contextual ELMo embedding model.'''

    emb_size: int = attr.ib()
    model: ElmoModel = attr.ib(factory=ElmoModel, repr=False)

    def __init__(self, emb_size: int = 1024, *, verbose: bool = True):
        '''
        ELMo model initialization.

        :param emb_size: supported embedding dimensionality.
        :param verbose: verbosity flag
        '''

        try:
            path = get_model_path(f'elmo{emb_size}')
            if not path.exists():
                raise OSError(f'Cannot load ELMo model from {path}')

            with catch_warnings():
                simplefilter(action='ignore', category=Warning)
                model = ElmoModel()
                model.load(str(path))

            self.__attrs_init__(verbose=verbose, logger=logger, emb_size=emb_size, model=model)
            self.log(f'Loaded ELMo-{emb_size} model.')
        except (FetchError, OSError):
            raise ModelError(
                'Cannot load ELMo model. Consider downloading '
                'required files using "semsim.download" '
                'or "semsim.download_all".'
            )

    def process_sentence(self, tokens: list[str]) -> Sequence[Embedding]:
        '''
        Produce embeddings for a list of tokens.

        :param tokens: list of tokens to produce embeddings for
        :return: sequence of embeddings, one per each token
        '''

        self.log('Generating sentence embeddings...')
        return self.process_batch([tokens])[0]

    def process_batch(self, tokens: list[list[str]]) -> Sequence[Sequence[Embedding]]:
        '''
        Produce embeddings for tokens, given a list of sentences split into tokens.

        :param tokens: list of lists of tokens, one list per sentence
        :return: sequence of embedding sequences, one sequence per sentence
        '''

        self.log('Generating batch embeddings...')
        with catch_warnings():
            simplefilter(action='ignore', category=Warning)
            vectors = self.model.get_elmo_vectors(tokens)
        return cast(Sequence[Sequence[Embedding]], vectors)


@attr.s(slots=True, kw_only=True, init=False)
class AutoEmbeddingModel(BaseEmbeddingModel):  # type: ignore
    '''Facade model for automatically instantiating an embedding model of any kind.'''

    embs_type: str | None = attr.ib()
    static_model: BaseStaticEmbeddingModel | None = attr.ib(repr=False)
    contextual_model: BaseContextualEmbeddingModel | None = attr.ib(repr=False)

    def __init__(
        self,
        embs_type: str | None = None,
        *,
        static_model: BaseStaticEmbeddingModel | None = None,
        contextual_model: BaseContextualEmbeddingModel | None = None,
        verbose: bool = True,
        **kwargs: Any,
    ):
        '''
        AutoEmbeddingModel initialization.

        :param embs_type: kind of embeddings to produce
        :param static_model: static embedding model to instantiate object with
        :param contextual_model: contextual embedding model to instantiate object with
        :param verbose: verbosity flag
        '''

        BaseEmbeddingModel.__init__(self, verbose=verbose, logger=logger)
        self.log('Detecting AutoEmbeddingModel type...')

        num_passed_args = sum(
            arg is not None
            for arg in (embs_type, static_model, contextual_model)
        )
        if num_passed_args > 1:
            raise ArgumentError(
                'You should specify at most one of "embs_type", "static_model" '
                'and "contextual_model" parameters.'
            )

        if static_model is not None:
            embs_type = 'custom_static_model'
        elif contextual_model is not None:
            embs_type = 'custom_contextual_model'
        elif embs_type is None:
            self.log(
                'Setting default AutoEmbeddingModel type to None. '
                'Make sure not to forget passing embedding type to AutoEmbeddingModel constructor.'
            )
        else:
            embs_type = embs_type.lower()
            if embs_type == 'cbow':
                static_model = CBOWModel(verbose=verbose, **kwargs)
            elif embs_type == 'elmo1024':
                contextual_model = ELMoModel(verbose=verbose, emb_size=1024, **kwargs)
            elif embs_type == 'elmo2048':
                contextual_model = ELMoModel(verbose=verbose, emb_size=2048, **kwargs)
            else:
                raise ArgumentError(
                    f'Unknown embedding model type: "{embs_type}".\n'
                    'Choose "cbow", "elmo1024", "elmo2048" or None.'
                )

        self.__attrs_init__(
            verbose=verbose,
            logger=logger,
            embs_type=embs_type,
            static_model=static_model,
            contextual_model=contextual_model,
        )

    def __attrs_post_init__(self) -> None:
        self.set_verbosity(self.verbose)

    def restore_embedding(self, node: Node) -> None:
        '''
        Restore static embedding for a given node.

        :param node: Node instance to restore embedding for
        :return: None
        '''

        if not self.is_static():
            raise ModelError('Cannot call "restore_embedding" on a non-static model.')
        if self.static_model is not None:
            self.static_model.restore_embedding(node)

    def process_sentence(self, tokens: list[str]) -> Sequence[Embedding]:
        '''
        Produce embeddings for a list of tokens.

        :param tokens: list of tokens to produce embeddings for
        :return: sequence of embeddings, one per each token
        '''

        if not self.is_contextual():
            raise ModelError('Cannot call "process_sentence" on a non-contextual model.')
        return self.contextual_model.process_sentence(tokens)  # type: ignore

    def process_batch(self, tokens: list[list[str]]) -> Sequence[Sequence[Embedding]]:
        '''
        Produce embeddings for tokens, given a list of sentences split into tokens.

        :param tokens: list of lists of tokens, one list per sentence
        :return: sequence of embedding sequences, one sequence per sentence
        '''

        if not self.is_contextual():
            raise ModelError('Cannot call "process_batch" on a non-contextual model.')
        return self.contextual_model.process_batch(tokens)  # type: ignore

    def set_verbosity(self, verbose: bool) -> None:
        '''
        Set verbosity value.

        :param verbose: verbosity flag
        :return: None
        '''

        super().set_verbosity(verbose)
        for model in (self.static_model, self.contextual_model):
            if model is not None:
                model.set_verbosity(verbose)

    def is_static(self) -> bool:
        '''
        Check whether a model generated static embeddings for each token.

        :return: True if model produces static vectors, False otherwise
        '''

        return self.contextual_model is None

    def is_contextual(self) -> bool:
        '''
        Check whether a model generated contextual (dynamic) embeddings for each token.

        :return: True if model produces contextual vectors, False otherwise
        '''

        return self.contextual_model is not None

    @property
    def emb_size(self) -> int:
        '''Embedding dimensionality.'''

        if self.is_static():
            return 0 if self.static_model is None else self.static_model.emb_size
        if self.is_contextual():
            return self.contextual_model.emb_size  # type: ignore
        raise ModelError('Cannot evaluate emb_size for empty AutoEmbeddingModel.')



'''Main high-level class for comparing texts.'''

import heapq
from pathlib import Path
from typing import Any

import attr
from attr import validators
from tqdm import tqdm

from .heuristic import BaseOrderHeuristic, PairwiseOrderHeuristic
from .kernel import BaseKernel, TABK
from ..logger import LoggerMixin, get_logger
from .representation import TextEngine, TextRepr
from .tree import Tree


__all__ = (
    'SemSimEngine',
)


logger = get_logger(__name__)


@attr.s(slots=True, kw_only=True, order=True)
class CompareResult:
    '''Structure representing result of tree pair comparison.'''

    lhs: Tree = attr.ib(order=False)
    rhs: Tree = attr.ib(order=False)
    score: float = attr.ib(validator=[validators.ge(0), validators.le(1)])


@attr.s(slots=True, kw_only=True, init=False)
class SemSimEngine(LoggerMixin):
    '''Main class for comparing texts.'''

    engine: TextEngine = attr.ib()
    kernel: BaseKernel = attr.ib(factory=TABK)
    heuristic: BaseOrderHeuristic = attr.ib()

    max_out_pairs: int = attr.ib(default=100, validator=validators.gt(0))
    max_process_pairs: int = attr.ib(default=10_000, validator=validators.gt(0))

    def __init__(self, *, verbose: bool = True, **kwargs: Any):
        '''
        SemSimEngine initialization.

        :param verbose: verbosity flag
        '''

        self.__attrs_init__(verbose=verbose, logger=logger, **kwargs)
        self.log('Successfully built SemSimEngine.\n')

    @engine.default
    def _(self) -> TextEngine:
        return TextEngine(verbose=self.verbose)

    @heuristic.default
    def _(self) -> BaseOrderHeuristic:
        return PairwiseOrderHeuristic(verbose=self.verbose)

    def __attrs_post_init__(self) -> None:
        self.set_verbosity(self.verbose)

    def set_verbosity(self, verbose: bool) -> None:
        '''
        Set verbosity value.

        :param verbose: verbosity flag
        :return: None
        '''

        super().set_verbosity(verbose)
        for attribute in (self.engine, self.heuristic):
            attribute.set_verbosity(verbose)

    def find_top_closest(
        self,
        first_source: str | Path,
        second_source: str | Path,
    ) -> list[CompareResult]:
        '''
        Get list of top sentence pairs sorted by score (descending), given text sources.

        Needed for tests compatibility, should not be used in reliable code.
        Use `find_top_closest_sentences` instead.

        :param first_source: first text source to compare, raw text or path to a file
        :param second_source: second text source to compare, raw text or path to a file
        :return: list of CompareResult instances, one per each sentence pair
        '''

        return self.find_top_closest_sentences(first_source, second_source)

    def find_top_closest_sentences(
        self,
        first_source: str | Path,
        second_source: str | Path,
    ) -> list[CompareResult]:
        '''
        Get list of top sentence pairs sorted by score (descending), given text sources.

        :param first_source: first text source to compare, raw text or path to a file
        :param second_source: second text source to compare, raw text or path to a file
        :return: list of CompareResult instances, one per each sentence pair
        '''

        first_repr = self.engine.process(first_source)
        second_repr = self.engine.process(second_source)
        return self.find_top_closest_trees(first_repr, second_repr)

    def find_top_closest_trees(
        self,
        first_repr: TextRepr,
        second_repr: TextRepr,
    ) -> list[CompareResult]:
        '''
        Get list of top sentence pairs sorted by score (descending), given tree pairs.

        :param first_repr: first text representation
        :param second_repr: second text representation
        :return: list of CompareResult instances, one per each sentence pair
        '''

        pairs = self.heuristic.select_pairs(first_repr, second_repr)
        total_pairs = min(len(first_repr) * len(second_repr), self.max_process_pairs)
        iter_pairs = tqdm(
            pairs,
            desc='Running kernel',
            total=total_pairs,
            disable=not self.verbose,
        )

        heap: list[CompareResult] = []
        for idx, (lhs, rhs) in enumerate(iter_pairs):
            if idx == self.max_process_pairs:
                break
            sim = self.kernel.compute(lhs, rhs)
            cmp = CompareResult(lhs=lhs, rhs=rhs, score=sim)
            if len(heap) == self.max_out_pairs:
                heapq.heappushpop(heap, cmp)
            else:
                heapq.heappush(heap, cmp)
        heap.sort(reverse=True)
        return heap



'''Sentence filters.'''

from abc import ABC, abstractmethod

import attr
from attr import validators

from .parser import split_conllu_sentence_tokens


__all__ = (
    'BaseSentenceFilter',
    'IdentitySentenceFilter',
    'MinLengthSentenceFilter',
)


@attr.s(slots=True, kw_only=True)
class BaseSentenceFilter(ABC):
    '''Base class for sentence filters.'''

    @abstractmethod
    def is_proper(self, conllu_sentence: str) -> bool:
        '''
        Check whether a sentence is appropriate for being processed.

        :param conllu_sentence: CoNLL-U data for a sentence to check
        :return: True if sentence is to be processed, False otherwise
        '''

        ...


@attr.s(slots=True, kw_only=True)
class IdentitySentenceFilter(BaseSentenceFilter):
    '''Lazy filter accepting every single sentence.'''

    def is_proper(self, conllu_sentence: str) -> bool:
        '''
        Check whether a sentence is appropriate for being processed.

        :param conllu_sentence: CoNLL-U data for a sentence to check
        :return: True if sentence is to be processed, False otherwise
        '''

        return True


@attr.s(slots=True, kw_only=True)
class MinLengthSentenceFilter(BaseSentenceFilter):
    '''Filter for ignoring too short sentences with few tokens.'''

    min_length: int = attr.ib(default=4, validator=validators.gt(0))

    def is_proper(self, conllu_sentence: str) -> bool:
        '''
        Check whether a sentence is appropriate for being processed.

        :param conllu_sentence: CoNLL-U data for a sentence to check
        :return: True if sentence is to be processed, False otherwise
        '''

        tokens = split_conllu_sentence_tokens(conllu_sentence)
        return len(tokens) >= self.min_length



'''SemSim exceptions.'''

__all__ = (
    'ArgumentError',
    'BaseSemSimException',
    'FetchError',
    'ModelError',
    'ParseError',
)


class BaseSemSimException(Exception):
    '''Base class for SemSim exceptions.'''

    pass


class ArgumentError(BaseSemSimException):
    '''Exception class for errors of passing wrong arguments.'''

    pass


class ModelError(BaseSemSimException):
    '''Exception class for model type errors.'''

    pass


class ParseError(BaseSemSimException):
    '''Exception class for text parsing errors.'''

    pass


class FetchError(BaseSemSimException):
    '''Exception class for errors while fetching data via HTTP GET-request.'''

    pass



'''Logging utilities.'''

from abc import ABC
from logging import Formatter, Logger, StreamHandler, getLogger

import attr


__all__ = (
    'LoggerMixin',

    'get_logger',
)


LOG_FORMAT = '%(asctime)s - [%(levelname)s] - %(message)s'


def get_stream_handler(level) -> StreamHandler:  # type: ignore
    '''
    Create stream handler for logger.

    :param level: logger level of verbosity
    :return: StreamHandler instance
    '''

    stream_handler = StreamHandler()
    stream_handler.setLevel('INFO')
    stream_handler.setFormatter(Formatter(LOG_FORMAT))
    return stream_handler


def get_logger(name: str, level: str = 'INFO') -> Logger:
    '''
    Create logger.

    :param name: name of logger
    :param level: logger level of verbosity
    :return: Logger instance
    '''

    logger = getLogger(name)
    logger.setLevel(level)
    logger.addHandler(get_stream_handler(level))
    return logger


@attr.s(slots=True, kw_only=True)
class LoggerMixin(ABC):
    '''Logging mixin class.'''

    verbose: bool = attr.ib(default=True, repr=False)
    logger: Logger | None = attr.ib(default=None, repr=False)

    def log(self, text: str) -> None:
        '''
        Write portion of text in a log.

        :param text: text to log
        :return: None
        '''

        if self.verbose and self.logger is not None:
            self.logger.info(text)

    def set_verbosity(self, verbose: bool) -> None:
        '''
        Set verbosity value.

        :param verbose: verbosity flag
        :return: None
        '''

        self.verbose = verbose

    def mute(self) -> None:
        '''
        Set verbosity value to False.

        :return: None
        '''

        self.set_verbosity(False)

    def unmute(self) -> None:
        '''
        Set verbosity value to True.

        :return: None
        '''

        self.set_verbosity(True)



'''Heuristics for selecting and ordering tree pairs to be compared.'''

from abc import ABC, abstractmethod
from operator import itemgetter
from typing import Any, Generator

import attr
from attr import validators

from .kernel import jaccard_index
from ..logger import LoggerMixin, get_logger
from .representation import TextRepr
from .tree import Tree


__all__ = (
    'BaseOrderHeuristic',
    'JaccardIndexOrderHeuristic',
    'PairwiseOrderHeuristic',
)


logger = get_logger(__name__)


PairOrder = Generator[tuple[Tree, Tree], None, None]


class BaseOrderHeuristic(LoggerMixin, ABC):
    '''Base class for pair order heuristics.'''

    def __init__(self, **kwargs: Any):
        '''Base order heuristic initialization.'''

        LoggerMixin.__init__(self, **kwargs)

    @abstractmethod
    def select_pairs(self, first_repr: TextRepr, second_repr: TextRepr) -> PairOrder:
        '''
        Select and order tree pairs to be processed.

        :param first_repr: first text representation
        :param second_repr: second text representation
        '''

        ...


@attr.s(slots=True, kw_only=True, init=False)
class PairwiseOrderHeuristic(BaseOrderHeuristic):
    '''Lazy heuristic producing all possible pairs of trees.'''

    def __init__(self, *, verbose: bool = True):
        '''
        Pairwise order heuristic initialization.

        :param verbose: verbosity flag
        '''

        BaseOrderHeuristic.__init__(self, verbose=verbose, logger=logger)

    def select_pairs(self, first_repr: TextRepr, second_repr: TextRepr) -> PairOrder:
        '''
        Select and order tree pairs to be processed.

        :param first_repr: first text representation
        :param second_repr: second text representation
        '''

        self.log('Running pair selecting heuristic...')
        for lhs in first_repr:
            for rhs in second_repr:
                yield lhs, rhs


@attr.s(slots=True, kw_only=True, init=False)
class JaccardIndexOrderHeuristic(BaseOrderHeuristic):
    '''Heuristic for comparing tree pairs with large Jaccard index of their token sets.'''

    min_jaccard: float = attr.ib(validator=validators.and_(validators.ge(0), validators.le(1)))
    use_lemma: bool = attr.ib()

    def __init__(self, *, min_jaccard: float = 0, use_lemma: bool = True, verbose: bool = True):
        '''
        Jaccard order heuristic initialization.

        :param min_jaccard: min value of Jaccard index to process tree pairs
        :param use_lemma: True if lemmas are used instead of tokens for Jaccard index evaluation
        :param verbose: verbosity flag
        '''

        self.__attrs_init__(
            verbose=verbose,
            logger=logger,
            min_jaccard=min_jaccard,
            use_lemma=use_lemma,
        )

    def _compute_jaccard(self, lhs: Tree, rhs: Tree) -> float:
        left_tokens = {node.lemma if self.use_lemma else node.token for node in lhs}
        right_tokens = {node.lemma if self.use_lemma else node.token for node in rhs}
        return jaccard_index(left_tokens, right_tokens)

    def select_pairs(self, first_repr: TextRepr, second_repr: TextRepr) -> PairOrder:
        '''
        Select and order tree pairs to be processed.

        :param first_repr: first text representation
        :param second_repr: second text representation
        '''

        self.log('Running pair selecting heuristic...')
        trees = [
            (lhs, rhs, self._compute_jaccard(lhs, rhs))
            for lhs in first_repr
            for rhs in second_repr
        ]
        trees.sort(key=itemgetter(2), reverse=True)

        for lhs, rhs, jac in trees:
            if jac < self.min_jaccard:
                return
            yield (lhs, rhs)



'''CLI tool implementation.'''

from pathlib import Path
from typing import Any

import click

from .. import (
    IdentitySentenceFilter, MinLengthSentenceFilter,
    PairwiseOrderHeuristic, JaccardIndexOrderHeuristic,
    SABK, TABK, MSK, CompositeKernel, AssignmentKernel, NeuralKernel,
    UDPipeSyntaxParser, AutoEmbeddingModel, CBOWModel,
    TextEngine, SemSimEngine,
    ArgumentError,
)
from ..algo.kernel import DISTANCES
from ..download import download_all, download as download_one


__all__ = (
    'compare',
    'download',
)


FILTER_TYPE_TO_CLS = {
    'id': IdentitySentenceFilter,
    'min-length': MinLengthSentenceFilter,
}

HEURISTIC_TYPE_TO_CLS = {
    'pairwise': PairwiseOrderHeuristic,
    'jaccard': JaccardIndexOrderHeuristic,
}

KERNEL_TYPE_TO_CLS = {
    'sabk': SABK,
    'tabk': TABK,
    'msk': MSK,
    'ck': CompositeKernel,
    'ak': AssignmentKernel,
    'nk': NeuralKernel,
    'composite': CompositeKernel,
    'assignment': AssignmentKernel,
    'neural': NeuralKernel,
}


def filter_none_values(dct: dict[str, Any]) -> dict[str, Any]:
    '''
    Filter out dictionary entries with None values.

    :param dct: dictionary to filter
    :return: filtered dictionary instance
    '''

    return {
        key: value
        for key, value in dct.items()
        if value is not None
    }


@click.group()
def cli() -> None:
    '''CLI group.'''

    pass


@cli.command()
@click.argument(  # texts to compare
    'first_src',
    type=click.Path(dir_okay=False, readable=True),
)
@click.argument(
    'second_src',
    type=click.Path(dir_okay=False, readable=True),
)
@click.option(  # syntax parser model
    '-s', '--syntax', 'syntax_path',
    type=click.Path(dir_okay=False, readable=True),
    help='Path to the UDPipe syntax model.',
)
@click.option(  # embedding model
    '-e', '--embs_type',
    type=click.Choice(['cbow', 'elmo1024', 'elmo2048', 'none'], case_sensitive=False),
    help='Embedding model type.',
)
@click.option(
    '--cbow_path',
    type=click.Path(dir_okay=False, readable=True),
    help='Path to a CBOW embedding model.',
)
@click.option(
    '-v/-q', '--verbose/--quiet',
    is_flag=True,
    default=True,
    show_default=True,
    help='Verbosity flag.',
)
@click.option(  # filter
    '-f', '--filter', 'filter_type',
    type=click.Choice(['id', 'min-length'], case_sensitive=False),
    default='min-length',
    show_default=True,
    help='Sentence filter type.',
)
@click.option(
    '--min-length',
    type=click.IntRange(min=1),
    help='Min length of sentences in tokens (for "min-length" filter only).',
)
@click.option(  # pair order heuristic
    '--heuristic', 'heuristic_type',
    type=click.Choice(['jaccard', 'pairwise'], case_sensitive=False),
    default='pairwise',
    show_default=True,
    help='Pair selection heuristic type.',
)
@click.option(
    '--min-jaccard',
    type=click.FloatRange(0, 1),
    help='Min Jaccard index value for filtering sentence pairs.',
)
@click.option(
    '--jaccard-lemma/--no-jaccard-lemma',
    is_flag=True,
    help='Use lemmas for jaccard index order heuristic.',
)
@click.option(  # kernel parameters
    '-k', '--kernel', 'kernel_type',
    type=click.Choice([
            'sabk', 'tabk', 'msk', 'ck', 'ak', 'nk',
            'composite', 'assignment', 'neural',
        ],
        case_sensitive=False,
    ),
    default='composite',
    show_default=True,
    help='Kernel type.',
)
@click.option(
    '--cmp-lemma/--no-cmp-lemma',
    is_flag=True,
    help='Compare lemmas instead of tokens in bigram kernels.',
)
@click.option(
    '--emb-metric',
    type=click.Choice(list(DISTANCES.keys()), case_sensitive=False),
    help='Embedding distance metric.',
)
@click.option(
    '-p',
    type=click.FloatRange(min=0, min_open=True),
    help='Parameter "p" for Minkowski distance.',
)
@click.option(
    '--theta',
    type=click.FloatRange(min=1),
    help='Parameter "theta" for dependency relation distance.',
)
@click.option(
    '--lambda', 'lambda_',
    type=click.FloatRange(min=0),
    help='Parameter "lambda" for dependency relation distance.',
)
@click.option(
    '--gamma',
    type=click.FloatRange(min=0, max=1, min_open=True),
    help='Parameter "gamma" for additive kernels.',
)
@click.option(
    '--idf-path',
    type=click.Path(dir_okay=False, readable=True),
    help='Path to an idf dictionary for TABK kernel.',
)
@click.option(
    '--unknown-idf',
    type=click.FloatRange(min=0),
    help='IDF value of unknown tokens for TABK kernel.',
)
@click.option(
    '--use-lemma/--no-use-lemma',
    is_flag=True,
    help='Use lemmas in TABK kernel.',
)
@click.option(
    '--alpha',
    type=click.FloatRange(min=0, max=1, min_open=True),
    help='Parameter "alpha" for MSK kernel.',
)
@click.option(
    '--nu',
    type=click.FloatRange(min=0, max=1, max_open=True),
    help='Parameter "nu" for MSK kernel.',
)
@click.option(
    '--beta',
    type=click.FloatRange(min=0, max=1),
    help='Parameter "beta" for CK kernel.',
)
@click.option(
    '--delta',
    type=click.FloatRange(min=0, max=1),
    help='Parameter "delta" for CK kernel.',
)
@click.option(
    '--gnn-path',
    type=click.Path(dir_okay=False, readable=True),
    help='Path to a PyTorch GNN model save (checkpoint).',
)
@click.option(
    '--gnn-type',
    type=click.Choice(['gcn', 'gat'], case_sensitive=False),
    help='GNN architecture type.',
)
@click.option(
    '--batch-size',
    type=click.IntRange(min=1),
    help='Batch size for GNN (in batched processing mode).',
)
@click.option(
    '--device',
    type=str,
    help='Type of device to load a GNN model and batches to.',
)
@click.option(  # additional SemSimEngine parameters
    '--max-out-pairs',
    type=click.IntRange(min=1),
    help='Max number of similar tree pairs to output.',
)
@click.option(
    '--max-process-pairs',
    type=click.IntRange(min=1),
    help='Max number of tree pairs to process.',
)
@click.option(
    '-o', '--output',
    type=click.Path(dir_okay=False, writable=True, readable=False),
    help='Output file path (stdout by default).',
)
def compare(
    first_src: str | Path,
    second_src: str | Path,

    syntax_path: str | Path | None,

    embs_type: str | None,
    cbow_path: str | Path | None,

    verbose: bool,

    filter_type: str,
    min_length: int | None,

    heuristic_type: str,
    min_jaccard: float | None,
    jaccard_lemma: bool | None,

    kernel_type: str,
    cmp_lemma: bool | None,
    emb_metric: str | None,
    p: float | None,
    theta: float | None,
    lambda_: float | None,
    gamma: float | None,
    idf_path: str | Path | None,
    unknown_idf: float | None,
    use_lemma: bool | None,
    alpha: float | None,
    nu: float | None,
    beta: float | None,
    delta: float | None,
    gnn_path: str | Path | None,
    gnn_type: str | None,
    batch_size: int | None,
    device: str | None,

    max_out_pairs: int | None,
    max_process_pairs: int | None,
    output: str | Path | None,
) -> None:
    '''CLI `compare` subcommand implementation.'''

    syntax_parser = UDPipeSyntaxParser(
        model_path=syntax_path,
    )

    if cbow_path is not None:
        if embs_type is not None:
            raise ArgumentError(
                'You should specify either "embs_type" or "cbow_path" '
                'parameter, but not both.'
            )
        cbow_model = CBOWModel(cbow_path)
        emb_model = AutoEmbeddingModel(static_model=cbow_model)
    else:
        if embs_type == 'none':
            embs_type = None
        emb_model = AutoEmbeddingModel(embs_type=embs_type)

    filter_cls = FILTER_TYPE_TO_CLS.get(filter_type)
    if filter_cls is None:
        raise ArgumentError(f'Unknown filter type "{filter_type}"')
    filter_params = filter_none_values({
        'min_length': min_length,
    })
    sentence_filter = filter_cls(**filter_params)

    engine = TextEngine(
        syntax_parser=syntax_parser,
        emb_model=emb_model,
        sentence_filter=sentence_filter,
    )

    heuristic_cls = HEURISTIC_TYPE_TO_CLS.get(heuristic_type)
    if heuristic_cls is None:
        raise ArgumentError(f'Unknown heuristic type "{heuristic_type}"')
    heuristic_params = filter_none_values({
        'min_jaccard': min_jaccard,
        'use_lemma': jaccard_lemma,
    })
    heuristic = heuristic_cls(**heuristic_params)

    idf = None
    if idf_path is not None:
        with open(idf_path, 'r', encoding='utf-8') as file:
            idf = {}
            for line in file:
                key, value = line.split(':')
                idf[key] = float(value)

    kernel_cls = KERNEL_TYPE_TO_CLS.get(kernel_type)
    if kernel_cls is None:
        raise ArgumentError(f'Unknown kernel type "{kernel_type}"')
    kernel_params = filter_none_values({
        'cmp_lemma': cmp_lemma,
        'emb_metric': emb_metric,
        'p': p,
        'theta': theta,
        'lambda_': lambda_,
        'gamma': gamma,
        'idf': idf,
        'unknown_idf': unknown_idf,
        'use_lemma': use_lemma,
        'alpha': alpha,
        'nu': nu,
        'beta': beta,
        'delta': delta,
        'gnn_path': gnn_path,
        'gnn_type': gnn_type,
        'batch_size': batch_size,
        'device': device,
    })
    kernel = kernel_cls(**kernel_params)

    semsim_engine_params = filter_none_values({
        'engine': engine,
        'kernel': kernel,
        'heuristic': heuristic,
        'verbose': verbose,
        'max_out_pairs': max_out_pairs,
        'max_process_pairs': max_process_pairs,
    })
    semsim_engine = SemSimEngine(**semsim_engine_params)

    results = semsim_engine.find_top_closest_sentences(Path(first_src), Path(second_src))

    if output is None:
        for idx, cmp in enumerate(results, start=1):
            print(f'Sentence pair #{idx}\n{cmp}', end='\n\n')
    else:
        with open(output, 'w', encoding='utf-8') as file:
            for idx, cmp in enumerate(results, start=1):
                print(f'Sentence pair #{idx}\n{cmp}', end='\n\n', file=file)


@cli.command()
@click.option(
    '-m', '--file', '--model',
    type=str,
    help='Model name',
)
@click.option(
    '-v/-q', '--verbose/--quiet',
    is_flag=True,
    default=True,
    show_default=True,
    help='Verbosity flag.',
)
@click.option(
    '-f', '--force',
    is_flag=True,
    default=False,
    show_default=True,
    help='Whether to force downloading already existing files.',
)
@click.option(
    '-a', '--all',
    is_flag=True,
    default=False,
    show_default=True,
    help='Pass this to download all built-in models at once.',
)
def download(
    file: str | None,
    verbose: bool,
    force: bool,
    all: bool,
) -> None:
    '''CLI `download` subcommand implementation.'''

    if all:
        if file is not None:
            raise ArgumentError('Cannot pass "file" argument along with "--all" flag.')
        download_all(verbose=verbose, force=force)
    elif file is None:
        raise ArgumentError('You should pass either "file" argument or "--all" flag, but not both.')
    else:
        download_one(file, verbose=verbose, force=force)


if __name__ == '__main__':
    cli()
